# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2025 The FreeCAD Project Association                    *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to upgrade objects by different methods.

See also the `downgrade` function.
"""
## @package downgrade
# \ingroup draftfunctions
# \brief Provides functions to upgrade objects by different methods.

import math
import re
import lazy_loader.lazy_loader as lz

import FreeCAD as App
from draftfunctions import draftify
from draftgeoutils.geometry import is_straight_line
from draftmake import make_block
from draftmake import make_wire
from draftutils import gui_utils
from draftutils import params
from draftutils import utils
from draftutils.groups import is_group
from draftutils.messages import _msg
from draftutils.translate import translate

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")
DraftGeomUtils = lz.LazyLoader("DraftGeomUtils", globals(), "DraftGeomUtils")
Arch = lz.LazyLoader("Arch", globals(), "Arch")

_DEBUG = False

## \addtogroup draftfunctions
# @{


def upgrade(objects, delete=False, force=None):
    """Upgrade the given objects.

    This is a counterpart to `downgrade`.

    Parameters
    ----------
    objects: Part::Feature or list
        A single object to upgrade or a list
        containing various such objects.

    delete: bool, optional
        It defaults to `False`.
        If it is `True`, the old objects are deleted, and only the resulting
        object is kept.

    force: str, optional
        It defaults to `None`.
        Its value can be used to force a certain method of upgrading.
        It can be any of: `'makeCompound'`, `'closeGroupWires'`,
        `'makeSolid'`, `'closeWire'`, `'turnToParts'`, `'makeFusion'`,
        `'makeShell'`, `'makeFaces'`, `'draftify'`, `'joinFaces'`,
        `'makeSketchFace'`, `'makeWires'`.

    Returns
    -------
    tuple
        A tuple containing two lists, a list of new objects
        and a list of objects to be deleted.

    See Also
    --------
    downgrade
    """

    # definitions of actions to perform

    def makeCompound(objects):
        """Return a compound object made from the given objects."""
        newobj = make_block.make_block(objects)
        format(objects[0], [newobj])
        add_to_parent(objects[0], [newobj])
        add_list.append(newobj)
        return True

    def closeGroupWires(groups):
        """Close every open wire in the given groups."""
        result = False
        for grp in groups:
            if any([closeWire(obj) for obj in grp.Group]):
                result = True
        return result

    def makeSolid(obj):
        """Turn an object into a solid, if possible."""
        if obj.Shape.Solids:
            return False
        try:
            solid = Part.makeSolid(obj.Shape)
        except Part.OCCError:
            return False
        if not solid.isClosed():
            return False
        newobj = doc.addObject("Part::Feature", "Solid")
        newobj.Shape = solid
        format(obj, [newobj])
        add_to_parent(obj, [newobj])
        add_list.append(newobj)
        delete_list.append(obj)
        return True

    def closeWire(obj):
        """Close a wire object, if possible."""
        if obj.Shape.Faces:
            return False
        if len(obj.Shape.Wires) != 1:
            return False
        if len(obj.Shape.Edges) == 1:
            return False
        if is_straight_line(obj.Shape):
            return False
        if utils.get_type(obj) == "Wire":
            obj.Closed = True
            return True
        wire = obj.Shape.Wires[0]
        if wire.isClosed():
            return False
        verts = wire.OrderedVertexes
        p0 = verts[0].Point
        p1 = verts[-1].Point
        edges = wire.Edges
        edges.append(Part.LineSegment(p1, p0).toShape())
        wire = Part.Wire(Part.__sortEdges__(edges))
        newobj = doc.addObject("Part::Feature", "Wire")
        newobj.Shape = wire
        format(obj, [newobj])
        add_to_parent(obj, [newobj])
        add_list.append(newobj)
        delete_list.append(obj)
        return True

    def turnToParts(meshes):
        """Turn given meshes to parts."""
        result = False
        for mesh in meshes:
            shp = Arch.getShapeFromMesh(mesh.Mesh)
            if shp:
                newobj = doc.addObject("Part::Feature", shp.ShapeType)
                newobj.Shape = shp
                format(mesh, [newobj])
                add_to_parent(mesh, [newobj])
                add_list.append(newobj)
                delete_list.append(mesh)
                result = True
        return result

    def makeFusion(objects):
        """Make a Draft or Part fusion between 2 given objects."""
        newobj = doc.addObject("Part::Fuse", "Fusion")
        newobj.Base = objects[0]
        newobj.Tool = objects[1]
        format(objects[0], [newobj])
        add_to_parent(objects[0], [newobj])
        add_list.append(newobj)
        return True

    def makeShell(objects):
        """Make a shell or compound with the given objects."""
        faces = []
        done_list = []
        for obj in objects:
            if obj.Shape.Faces:
                faces.extend(obj.Shape.Faces)
                done_list.append(obj)
        if not faces:
            return None
        shp = Part.makeShell(faces)
        if shp.isNull():
            return None
        newobj = doc.addObject("Part::Feature", shp.ShapeType)
        newobj.Shape = shp
        # Format before applying diffuse color:
        format(done_list[0], [newobj])
        add_to_parent(done_list[0], [newobj])
        add_list.append(newobj)
        delete_list.extend(done_list)

        if App.GuiUp and params.get_param("preserveFaceColor"):
            # Must happen after add_to_parent for correct CenterOfMass.
            colors = gui_utils.get_diffuse_color(done_list)
            if len(faces) != len(colors):
                newobj.ViewObject.DiffuseColor = [colors[0]]
            else:
                # The ordering of shp.Faces may be different. Since we cannot
                # compare via hashCode(), we have to iterate and use different
                # criteria to find the correct color.
                old_data = []
                for face, color in zip(faces, colors):
                    old_data.append([face.Area, face.CenterOfMass, color])
                new_colors = []
                for new_face in shp.Faces:
                    new_area = new_face.Area
                    new_cen = new_face.CenterOfMass
                    for old_area, old_cen, old_color in old_data:
                        if math.isclose(new_area, old_area, abs_tol=1e-7) and new_cen.isEqual(
                            old_cen, 1e-7
                        ):
                            new_colors.append(old_color)
                            break
                newobj.ViewObject.DiffuseColor = new_colors

        if params.get_param("preserveFaceNames"):
            firstName = done_list[0].Label
            nameNoTrailNumbers = re.sub(r"\d+$", "", firstName)
            newobj.Label = "{} {}".format(newobj.Label, nameNoTrailNumbers)

        return newobj

    def joinFaces(objects):
        """Make one big face from the given objects, if possible."""
        faces = []
        done_list = []
        for obj in objects:
            if obj.Shape.Faces:
                faces.extend(obj.Shape.Faces)
                done_list.append(obj)
        if not faces:
            return False
        if not DraftGeomUtils.is_coplanar(faces, 1e-3):
            return False
        fuse_face = faces.pop(0)
        for face in faces:
            fuse_face = fuse_face.fuse(face)
        face = DraftGeomUtils.concatenate(fuse_face)
        # check if concatenate failed
        if face.isEqual(fuse_face):
            return False
        # several coplanar and non-curved faces, they can become a Draft Wire
        if len(face.Wires) == 1 and not DraftGeomUtils.hasCurves(face):
            newobj = make_wire.make_wire(face.Wires[0], closed=True, face=True)
        # if not possible, we do a non-parametric union
        else:
            newobj = doc.addObject("Part::Feature", "Union")
            newobj.Shape = face
        format(done_list[0], [newobj])
        add_to_parent(done_list[0], [newobj])
        add_list.append(newobj)
        delete_list.extend(done_list)
        return True

    def makeSketchFace(obj):
        """Make a face from a sketch."""
        face = Part.makeFace(obj.Shape.Wires, "Part::FaceMakerBullseye")
        if not face:
            return False
        newobj = doc.addObject("Part::Feature", "Face")
        newobj.Shape = face
        format(obj, [newobj])
        add_to_parent(obj, [newobj])
        add_list.append(newobj)
        delete_list.append(obj)
        return True

    def makeFaces(objects):
        """Make a face from every closed wire in the given objects."""
        result = False
        for obj in objects:
            new_list = []
            for wire in obj.Shape.Wires:
                try:
                    face = Part.Face(wire)
                except Part.OCCError:
                    continue
                newobj = doc.addObject("Part::Feature", "Face")
                newobj.Shape = face
                new_list.append(newobj)
            if not new_list:
                continue
            format(obj, new_list)
            add_to_parent(obj, new_list)
            add_list.extend(new_list)
            delete_list.append(obj)
            result = True
        return result

    def makeWires(objects):
        """Join edges in the given objects into wires."""
        edges = []
        done_list = []
        for obj in objects:
            if obj.Shape.Edges:
                edges.extend(obj.Shape.Edges)
                done_list.append(obj)
        if not edges:
            return False
        try:
            sorted_edges = Part.sortEdges(edges)
            if _DEBUG:
                for cluster in sorted_edges:
                    for edge in cluster:
                        print("Curve: {}".format(edge.Curve))
                        print(
                            "first: {}, last: {}".format(
                                edge.Vertexes[0].Point, edge.Vertexes[-1].Point
                            )
                        )
            wires = [Part.Wire(cluster) for cluster in sorted_edges]
        except Part.OCCError:
            return False
        if len(objects) > 1 and len(wires) == len(objects):
            # we still have the same number of objects, we actually didn't join anything!
            return False
        new_list = []
        for wire in wires:
            newobj = doc.addObject("Part::Feature", "Wire")
            newobj.Shape = wire
            new_list.append(newobj)
        # We don't know which wire came from which obj, we format them the same:
        format(done_list[0], new_list)
        add_to_parent(done_list[0], new_list)
        add_list.extend(new_list)
        delete_list.extend(done_list)
        return True

    def _draftify(obj):
        """Wrapper for draftify."""
        new_list = draftify.draftify(obj, delete=False)
        if not new_list:
            return False
        if not isinstance(new_list, list):
            new_list = [new_list]
        format(obj, new_list)
        add_to_parent(obj, new_list)
        add_list.extend(new_list)
        delete_list.append(obj)
        return True

    # helper functions (same as in downgrade.py)

    def get_parent(obj):
        # Problem with obj.getParent():
        # https://github.com/FreeCAD/FreeCAD/issues/19600
        parent = obj.getParentGroup()
        if parent is not None:
            return parent
        return obj.getParentGeoFeatureGroup()

    def can_be_deleted(obj):
        if not obj.InList:
            return True
        for other in obj.InList:
            if is_group(other):
                continue
            if other.TypeId == "App::Part":
                continue
            return False
        return True

    def delete_object(obj):
        if utils.is_deleted(obj):
            return
        parent = get_parent(obj)
        if parent is not None and parent.TypeId == "PartDesign::Body":
            obj = parent
        if not can_be_deleted(obj):
            # Make obj invisible instead:
            obj.Visibility = False
            return
        if obj.TypeId == "PartDesign::Body":
            obj.removeObjectsFromDocument()
        doc.removeObject(obj.Name)

    def add_to_parent(obj, new_list):
        parent = get_parent(obj)
        if parent is None:
            if doc.getObject("Draft_Construction"):
                # This cludge is required because the make_* commands may
                # put new objects in the construction group.
                constr_group = doc.getObject("Draft_Construction")
                for newobj in new_list:
                    constr_group.removeObject(newobj)
            return
        if parent.TypeId == "PartDesign::Body":
            # We don't add to a PD Body. We process its placement and
            # add to its parent instead.
            for newobj in new_list:
                newobj.Placement = parent.Placement.multiply(newobj.Placement)
            add_to_parent(parent, new_list)
            return
        for newobj in new_list:
            # Using addObject is different from just changing the Group property.
            # With addObject the object will be added to the parent group, but if
            # that is a normal group, also to that group's parent GeoFeatureGroup,
            # if available.
            parent.addObject(newobj)

    def format(obj, new_list):
        for newobj in new_list:
            gui_utils.format_object(newobj, obj, ignore_construction=True)

    doc = App.ActiveDocument
    add_list = []
    delete_list = []
    result = False

    if not isinstance(objects, list):
        objects = [objects]
    if not objects:
        return add_list, delete_list

    # analyzing objects
    faces = []
    wires = []
    openwires = []
    facewires = []
    edges = []
    loneedges = []
    groups = []
    meshes = []
    parts = []

    for obj in objects:
        if obj.TypeId == "App::DocumentObjectGroup":
            groups.append(obj)
        elif hasattr(obj, "Shape"):
            parts.append(obj)
            faces.extend(obj.Shape.Faces)
            wires.extend(obj.Shape.Wires)
            edges.extend(obj.Shape.Edges)
            for face in obj.Shape.Faces:
                facewires.extend(face.Wires)
            wirededges = []
            for wire in obj.Shape.Wires:
                if len(wire.Edges) > 1:
                    for edge in wire.Edges:
                        wirededges.append(edge.hashCode())
                if not wire.isClosed():
                    openwires.append(wire)
            for edge in obj.Shape.Edges:
                if not edge.hashCode() in wirededges and not edge.isClosed():
                    loneedges.append(edge)
        elif obj.isDerivedFrom("Mesh::Feature"):
            meshes.append(obj)
    objects = parts

    if _DEBUG:
        print("objects: {}, edges: {}".format(objects, edges))
        print("wires: {}, openwires: {}".format(wires, openwires))
        print("faces: {}".format(faces))
        print("groups: {}".format(groups))
        print("facewires: {}, loneedges: {}".format(facewires, loneedges))

    if not (groups or objects or meshes):
        result = False

    elif force:
        if force == "closeGroupWires":
            result = closeGroupWires(groups)
        elif force == "turnToParts":
            result = turnToParts(meshes)
        else:
            # functions that work on a single object:
            single_funcs = {
                "closeWire": closeWire,
                "draftify": _draftify,
                "makeSketchFace": makeSketchFace,
                "makeSolid": makeSolid,
            }
            # functions that work on multiple objects:
            multi_funcs = {
                "joinFaces": joinFaces,
                "makeCompound": makeCompound,
                "makeFaces": makeFaces,
                "makeFusion": makeFusion,
                "makeShell": makeShell,
                "makeWires": makeWires,
            }
            if force in single_funcs:
                result = any([single_funcs[force](obj) for obj in objects])
            elif force in multi_funcs:
                result = multi_funcs[force](objects)
            else:
                _msg(translate("draft", "Upgrade: Unknown force method:") + " " + force)
                result = False

    # if we have a group: close each wire inside
    elif groups:
        result = closeGroupWires(groups)
        if result:
            _msg(translate("draft", "Found groups: closing open wires inside"))

    # if we have meshes, we try to turn them into shapes
    elif meshes:
        result = turnToParts(meshes)
        if result:
            _msg(translate("draft", "Found meshes: turning them into Part shapes"))

    else:
        # checking faces coplanarity
        # The precision needed in Part.makeFace is 1e-7. Here we use a
        # higher value to let that function throw the exception when
        # joinFaces is called if the precision is insufficient.
        if faces:
            faces_coplanarity = DraftGeomUtils.is_coplanar(faces, 1e-3)

        parent = get_parent(objects[0])
        same_parent = True
        same_parent_type = getattr(parent, "TypeId", "")  # "" for global space.
        if len(objects) > 1:
            for obj in objects[1:]:
                if get_parent(obj) != parent:
                    same_parent = False
                    same_parent_type = None
                    break

        # we have only faces
        if faces and len(facewires) == len(wires) and not openwires and not loneedges:

            # we have one shell: we try to make a solid
            # this also handles PD Bodies and PD features with solids (result will be False)
            if len(objects) == 1 and len(faces) > 3 and not faces_coplanarity:
                result = makeSolid(objects[0])
                if result:
                    _msg(translate("draft", "Found 1 solidifiable object: solidifying it"))

            # we have exactly 2 objects: we fuse them
            elif (
                len(objects) == 2
                and not faces_coplanarity
                and same_parent
                and same_parent_type != "PartDesign::Body"
            ):
                result = makeFusion(objects)
                if result:
                    _msg(translate("draft", "Found 2 objects: fusing them"))

            # we have many separate faces: we try to make a shell or compound
            elif (
                len(objects) > 1
                and len(faces) > 1
                and same_parent
                and same_parent_type != "PartDesign::Body"
            ):
                result = makeShell(objects)
                if result:
                    _msg(
                        translate(
                            "draft", "Found several objects: creating a " + result.Shape.ShapeType
                        )
                    )

            # we have faces: we try to join them if they are coplanar
            elif len(objects) == 1 and len(faces) > 1 and faces_coplanarity:
                result = joinFaces(objects)
                if result:
                    _msg(
                        translate(
                            "draft", "Found object with several coplanar faces: refining them"
                        )
                    )

            # only one object: if not parametric, we "draftify" it
            elif (
                len(objects) == 1
                and not objects[0].isDerivedFrom("Part::Part2DObjectPython")
                and not utils.get_type(objects[0]) in ["BezCurve", "BSpline", "Wire"]
            ):
                result = _draftify(objects[0])
                if result:
                    _msg(
                        translate(
                            "draft",
                            "Found 1 non-parametric object: replacing it with a Draft object",
                        )
                    )

        # in the following cases there are no faces
        elif not faces:

            # we have only closed wires
            if wires and not openwires and not loneedges:
                # we have a sketch: extract a face
                if len(objects) == 1 and objects[0].isDerivedFrom("Sketcher::SketchObject"):
                    result = makeSketchFace(objects[0])
                    if result:
                        _msg(
                            translate(
                                "draft", "Found 1 closed sketch object: creating a face from it"
                            )
                        )
                # only closed wires
                else:
                    result = makeFaces(objects)
                    if result:
                        _msg(translate("draft", "Found closed wires: creating faces"))

            # wires or edges: we try to join them
            elif len(objects) > 1 and len(edges) > 1 and same_parent:
                result = makeWires(objects)
                if result:
                    _msg(translate("draft", "Found several wires or edges: wiring them"))
                else:
                    result = makeCompound(objects)
                    if result:
                        _msg(
                            translate(
                                "draft", "Found several non-treatable objects: creating compound"
                            )
                        )

            # special case, we have only one open wire. We close it, unless it has only 1 edge!
            elif len(objects) == 1 and len(openwires) == 1:
                result = closeWire(objects[0])
                if result:
                    _msg(translate("draft", "Found 1 open wire: closing it"))

            # only one object: if not parametric, we "draftify" it
            elif (
                len(objects) == 1
                and len(edges) == 1
                and not objects[0].isDerivedFrom("Part::Part2DObjectPython")
                and not utils.get_type(objects[0]) in ["BezCurve", "BSpline", "Wire"]
            ):
                edge_type = DraftGeomUtils.geomType(objects[0].Shape.Edges[0])
                # currently only support Line and Circle
                if edge_type in ("Line", "Circle"):
                    result = _draftify(objects[0])
                    if result:
                        _msg(
                            translate(
                                "draft",
                                "Found 1 non-parametric object: replacing it with a Draft object",
                            )
                        )

            # only points, no edges
            elif not edges and len(objects) > 1:
                result = makeCompound(objects)
                if result:
                    _msg(translate("draft", "Found points: creating compound"))

        # all other cases, if more than 1 object, make a compound
        elif len(objects) > 1:
            result = makeCompound(objects)
            if result:
                _msg(translate("draft", "Found several non-treatable objects: creating compound"))

    # no result has been obtained
    if not result:
        _msg(translate("draft", "Unable to upgrade these objects"))

    if delete:
        for obj in delete_list:
            delete_object(obj)

    gui_utils.select(add_list)
    return add_list, delete_list


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Widget for editing a ToolBit object."""

from typing import Optional
from PySide import QtGui, QtCore
import FreeCAD
import FreeCADGui
from ...shape.ui.shapewidget import ShapeWidget
from ...docobject.ui import DocumentObjectEditorWidget
from ..models.base import ToolBit

translate = FreeCAD.Qt.translate


class ToolBitPropertiesWidget(QtGui.QWidget):
    """
    A composite widget for editing the properties and shape of a ToolBit.
    """

    # Signal emitted when the toolbit data has been modified
    toolBitChanged = QtCore.Signal()
    toolNoChanged = QtCore.Signal(int)

    def __init__(
        self,
        toolbit: Optional[ToolBit] = None,
        tool_no: Optional[int] = None,
        parent=None,
        icon: bool = True,
    ):
        super().__init__(parent)
        self._toolbit = None
        self._show_shape = icon
        self._tool_no = tool_no

        # Set schema to user preference if no document is open
        # TODO: Add a preference for toolbit unit schema.
        # We probably want to look at making it possible to have a toolbit be metric
        # or imperial regardless of document settings / or user preferences, but for now this is sufficient.
        if FreeCAD.ActiveDocument is None:
            pref_schema = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Units").GetInt(
                "UserSchema", 0
            )
            FreeCAD.Units.setSchema(pref_schema)

        # UI Elements
        self._label_edit = QtGui.QLineEdit()
        self._id_label = QtGui.QLabel()  # Read-only ID
        self._id_label.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)

        theicon = toolbit.get_icon() if toolbit else None
        abbr = theicon.abbreviations if theicon else {}
        self._property_editor = DocumentObjectEditorWidget(property_suffixes=abbr)
        self._property_editor.setSizePolicy(
            QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding
        )
        self._shape_widget = None  # Will be created in load_toolbit

        # Layout
        toolbit_group_box = QtGui.QGroupBox(translate("CAM", "Toolbit"))
        form_layout = QtGui.QFormLayout(toolbit_group_box)
        form_layout.addRow(translate("CAM", "Label:"), self._label_edit)
        form_layout.addRow(translate("CAM", "ID:"), self._id_label)

        # Optional tool number edit field.
        self._tool_no_edit = QtGui.QSpinBox()
        self._tool_no_edit.setMinimum(1)
        self._tool_no_edit.setMaximum(99999999)
        if tool_no is not None:
            form_layout.addRow(translate("CAM", "Tool Number:"), self._tool_no_edit)

        main_layout = QtGui.QVBoxLayout(self)
        main_layout.addWidget(toolbit_group_box)

        properties_group_box = QtGui.QGroupBox(FreeCAD.Qt.translate("CAM", "Properties"))
        properties_group_box.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)
        properties_layout = QtGui.QVBoxLayout(properties_group_box)
        properties_layout.setSpacing(5)
        properties_layout.addWidget(self._property_editor)

        # Ensure the layout expands horizontally
        properties_layout.setAlignment(QtCore.Qt.AlignLeft | QtCore.Qt.AlignTop)

        # Set stretch factor to make property editor expand
        properties_layout.setStretchFactor(self._property_editor, 1)

        main_layout.addWidget(properties_group_box)
        main_layout.addStretch(1)

        # Layout for centering the shape widget (created later)
        self._shape_display_layout = QtGui.QHBoxLayout()
        self._shape_display_layout.addStretch(1)

        # Placeholder for the widget
        self._shape_display_layout.addStretch(1)
        main_layout.addLayout(self._shape_display_layout)

        # Connections
        self._label_edit.editingFinished.connect(self._on_label_changed)
        self._tool_no_edit.valueChanged.connect(self._on_tool_no_changed)
        self._property_editor.propertyChanged.connect(self.toolBitChanged)

        if toolbit:
            self.load_toolbit(toolbit)

    def _on_label_changed(self):
        """Update the toolbit's label when the line edit changes."""
        if self._toolbit and self._toolbit.obj:
            new_label = self._label_edit.text()
            if self._toolbit.obj.Label != new_label:
                self._toolbit.obj.Label = new_label
                self.toolBitChanged.emit()

    def _on_tool_no_changed(self, value):
        """Update the tool number when the line edit changes."""
        if self._tool_no != value:
            self._tool_no = value
            self.toolNoChanged.emit(value)

    def load_toolbit(self, toolbit: ToolBit):
        """Load a ToolBit object into the editor."""
        self._toolbit = toolbit
        if not self._toolbit or not self._toolbit.obj:
            # Clear or disable fields if toolbit is invalid
            self._label_edit.clear()
            self._label_edit.setEnabled(False)
            self._id_label.clear()
            self._tool_no_edit.clear()
            self._property_editor.setObject(None)
            # Clear existing shape widget if any
            if self._shape_widget:
                self._shape_display_layout.removeWidget(self._shape_widget)
                self._shape_widget.deleteLater()
                self._shape_widget = None
            self._tool_no_edit.setValue(1)
            self.setEnabled(False)
            return

        self.setEnabled(True)
        self._label_edit.setEnabled(True)
        self._label_edit.setText(self._toolbit.obj.Label)
        self._id_label.setText(self._toolbit.get_id())
        self._tool_no_edit.setValue(int(self._tool_no or 1))

        # Get properties and suffixes
        props_to_show = self._toolbit._get_props(("Shape", "Attributes"))
        icon = self._toolbit._tool_bit_shape.get_icon()
        suffixes = icon.abbreviations if icon else {}
        self._property_editor.setObject(self._toolbit.obj)
        self._property_editor.setPropertiesToShow(props_to_show, suffixes)

        # Clear old shape widget and create/add new one if shape exists
        if self._shape_widget:
            self._shape_display_layout.removeWidget(self._shape_widget)
            self._shape_widget.deleteLater()
            self._shape_widget = None

        if self._show_shape and self._toolbit._tool_bit_shape:
            self._shape_widget = ShapeWidget(shape=self._toolbit._tool_bit_shape, parent=self)
            self._shape_widget.setMinimumSize(200, 150)
            # Insert into the middle slot of the HBox layout
            self._shape_display_layout.insertWidget(1, self._shape_widget)

    def save_toolbit(self):
        """
        Applies changes from the editor widgets back to the ToolBit object.
        Note: Most changes are applied via signals, but this can be called
        for explicit save actions.
        """
        # Ensure label is updated if focus is lost without pressing Enter
        self._on_label_changed()

        # No need to explicitly save the toolbit object itself here,
        # as properties were modified directly on toolbit.obj


class ToolBitEditorPanel(QtGui.QWidget):
    """
    A widget for editing a ToolBit object, wrapping ToolBitEditorWidget
    and providing standard dialog buttons.
    """

    # Signals
    accepted = QtCore.Signal(ToolBit)
    rejected = QtCore.Signal()
    toolBitChanged = QtCore.Signal()  # Re-emit signal from inner widget

    def __init__(self, toolbit: ToolBit | None = None, parent=None):
        super().__init__(parent)

        # Create the main editor widget
        self._editor_widget = ToolBitPropertiesWidget(toolbit, self)

        # Create the button box
        buttons = QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Cancel
        self._button_box = QtGui.QDialogButtonBox(buttons)

        # Connect button box signals to custom signals
        self._button_box.accepted.connect(self._accepted)
        self._button_box.rejected.connect(self.rejected.emit)

        # Layout
        main_layout = QtGui.QVBoxLayout(self)
        main_layout.addWidget(self._editor_widget)
        main_layout.addWidget(self._button_box)

        # Connect the toolBitChanged signal from the inner widget
        self._editor_widget.toolBitChanged.connect(self.toolBitChanged)

    def _accepted(self):
        self.accepted.emit(self._editor_widget._toolbit)

    def load_toolbit(self, toolbit: ToolBit):
        """Load a ToolBit object into the editor."""
        self._editor_widget.load_toolbit(toolbit)

    def save_toolbit(self):
        """Applies changes from the editor widgets back to the ToolBit object."""
        self._editor_widget.save_toolbit()


class ToolBitEditor(QtGui.QWidget):
    """
    A widget for editing a ToolBit object, wrapping ToolBitEditorWidget
    and providing standard dialog buttons.
    """

    # Signals
    toolBitChanged = QtCore.Signal()  # Re-emit signal from inner widget

    def __init__(
        self,
        toolbit: ToolBit,
        tool_no: Optional[int] = None,
        parent=None,
        icon: bool = False,
    ):
        super().__init__(parent)
        self.form = FreeCADGui.PySideUic.loadUi(":/panels/ToolBitEditor.ui")

        self.toolbit = toolbit
        self.tool_no = tool_no
        self.default_title = self.form.windowTitle()

        # Get first tab from the form, add the shape widget to the right.
        tool_tab_layout = self.form.toolTabLayout

        # Create a horizontal layout for the tab content
        tab_content_layout = QtGui.QHBoxLayout()

        # Add tool properties editor to the left with stretch
        props = ToolBitPropertiesWidget(toolbit, tool_no, self, icon=icon)

        # Wrap properties in a scroll area for vertical scrolling
        scroll_area = QtGui.QScrollArea()
        scroll_area.setWidget(props)
        scroll_area.setWidgetResizable(True)
        scroll_area.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        scroll_area.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        scroll_area.setMinimumHeight(550)  # Set minimum height for the scroll area
        scroll_area.setMaximumHeight(600)  # Set maximum height to prevent excessive growth

        tab_content_layout.addWidget(scroll_area, 1)

        # Add shape widget to the right without stretch
        widget = ShapeWidget(toolbit._tool_bit_shape)
        tab_content_layout.addWidget(widget, 0)

        # Add the horizontal layout to the tab layout
        tool_tab_layout.addLayout(tab_content_layout)

        props.toolBitChanged.connect(self._update)
        props.toolNoChanged.connect(self._on_tool_no_changed)

        self.form.tabWidget.setCurrentIndex(0)
        self.form.tabWidget.currentChanged.connect(self._on_tab_switched)

        # Hide second tab (tool notes) for now.
        self.form.tabWidget.setTabVisible(1, False)

        # Feeds & Speeds
        self.feeds_tab_idx = None
        """
        TODO: disabled for now.
        if tool.supports_feeds_and_speeds():
            label = translate('CAM', 'Feeds && Speeds')
            self.feeds = FeedsAndSpeedsWidget(db, serializer, tool, parent=self)
            self.feeds_tab_idx = self.form.tabWidget.insertTab(1, self.feeds, label)
        else:
            self.feeds = None
            self.feeds_tab_idx = None

        self.form.lineEditCoating.setText(toolbit.get_coating())
        self.form.lineEditCoating.textChanged.connect(toolbit.set_coating)
        self.form.lineEditHardness.setText(toolbit.get_hardness())
        self.form.lineEditHardness.textChanged.connect(toolbit.set_hardness)
        self.form.lineEditMaterials.setText(toolbit.get_materials())
        self.form.lineEditMaterials.textChanged.connect(toolbit.set_materials)
        self.form.lineEditSupplier.setText(toolbit.get_supplier())
        self.form.lineEditSupplier.textChanged.connect(toolbit.set_supplier)
        self.form.plainTextEditNotes.setPlainText(tool.get_notes())
        self.form.plainTextEditNotes.textChanged.connect(self._on_notes_changed)
        """

    def _update(self):
        title = self.default_title
        tool_name = self.toolbit.label
        if tool_name:
            title = "{} - {}".format(tool_name, title)
        self.form.setWindowTitle(title)

    def _on_tab_switched(self, index):
        if index == self.feeds_tab_idx:
            self.feeds.update()

    def _on_notes_changed(self):
        self.toolbit.set_notes(self.form.plainTextEditNotes.toPlainText())

    def _on_tool_no_changed(self, value):
        self.tool_no = value

    def get_tool_no(self):
        return self.tool_no

    def show(self):
        return self.form.exec_()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2025 Furgo                                              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Arch
import Part
from bimtests import TestArchBase


class TestArchRebar(TestArchBase.TestArchBase):

    def _create_sketch(self, add_line=True, length=500.0):
        """Helper function to create a basic sketch."""
        sketch = self.document.addObject("Sketcher::SketchObject")
        if add_line:
            sketch.addGeometry(
                Part.LineSegment(FreeCAD.Vector(0, 0, 0), FreeCAD.Vector(length, 0, 0)), False
            )
        self.document.recompute()
        return sketch

    def _create_host_structure(self, length=1000.0, width=200.0, height=300.0):
        """Helper function to create a basic host structure."""
        host = Arch.makeStructure(length=length, width=width, height=height)
        self.document.recompute()
        return host

    def test_makeRebar_default_name(self):
        """Test makeRebar creates rebar with default 'Rebar' Label.

        See https://github.com/FreeCAD/FreeCAD/issues/21670
        """
        sketch = self._create_sketch()
        rebar = Arch.makeRebar(sketch=sketch, diameter=6)
        self.document.recompute()

        self.assertIsNotNone(rebar, "makeRebar failed to create a rebar object.")
        self.assertEqual(rebar.Label, "Rebar", "Rebar default label should be 'Rebar'.")
        self.assertTrue(
            rebar.Name.startswith("Rebar"), "Rebar internal name should start with 'Rebar'."
        )

    def test_makeRebar_custom_name(self):
        """Test makeRebar with a custom name sets Label and Mark correctly."""
        sketch = self._create_sketch()
        custom_name = "MyCustomRebar"
        rebar = Arch.makeRebar(sketch=sketch, diameter=16, amount=5, name=custom_name)
        self.document.recompute()

        self.assertIsNotNone(rebar, "makeRebar failed to create a rebar object.")
        self.assertEqual(rebar.Label, custom_name, "Rebar label is incorrect.")
        self.assertEqual(rebar.Mark, custom_name, "Rebar mark should match label by default.")

    def test_makeRebar_with_sketch_and_host(self):
        """Test makeRebar with both sketch and host object links them correctly."""
        sketch_len = 800.0
        host_len = 1000.0
        sketch = self._create_sketch(length=sketch_len)
        host = self._create_host_structure(length=host_len)

        rebar = Arch.makeRebar(baseobj=host, sketch=sketch, diameter=10, amount=3, offset=20)
        self.document.recompute()

        self.assertIsNotNone(rebar, "makeRebar failed.")
        self.assertEqual(rebar.Host, host, "Rebar.Host is not set correctly.")
        self.assertEqual(rebar.Base, sketch, "Rebar.Base (sketch) is not set correctly.")
        self.assertAlmostEqual(rebar.Diameter.Value, 10.0, delta=1e-9)
        self.assertEqual(rebar.Amount, 3)
        self.assertAlmostEqual(rebar.OffsetStart.Value, 20.0, delta=1e-9)

        self.assertTrue(hasattr(rebar, "Shape"), "Rebar should have a Shape attribute.")
        self.assertIsNotNone(rebar.Shape, "Rebar.Shape should not be None after recompute.")
        self.assertTrue(rebar.Shape.isValid(), "Rebar.Shape should be valid.")
        self.assertGreater(rebar.Shape.Length, 0, "Rebar shape seems to have no length.")
        self.assertGreater(
            rebar.TotalLength.Value, 0, "Rebar total length should be greater than 0."
        )

    def test_makeRebar_with_sketch_only(self):
        """Test makeRebar with only sketch results in no host and valid geometry."""
        sketch = self._create_sketch()
        rebar = Arch.makeRebar(sketch=sketch, diameter=8)
        self.document.recompute()

        self.assertIsNotNone(rebar, "makeRebar failed.")
        self.assertIsNone(rebar.Host, "Rebar.Host should be None.")
        self.assertEqual(rebar.Base, sketch, "Rebar.Base (sketch) is not set correctly.")
        self.assertAlmostEqual(rebar.Diameter.Value, 8.0, delta=1e-9)
        self.assertIsNotNone(rebar.Shape)
        self.assertTrue(rebar.Shape.isValid())
        self.assertGreater(rebar.Shape.Length, 0)

    def test_makeRebar_with_baseobj_only_shape_conversion(self):
        """Test makeRebar with baseobj only copies shape."""
        host_shape_provider = self._create_host_structure()
        rebar = Arch.makeRebar(baseobj=host_shape_provider)
        self.document.recompute()

        self.assertIsNotNone(rebar, "makeRebar failed.")
        self.assertIsNone(rebar.Host, "Rebar.Host should be None in this mode.")
        self.assertIsNone(rebar.Base, "Rebar.Base should be None in this mode.")
        self.assertIsNotNone(rebar.Shape, "Rebar.Shape should be set from baseobj.")
        self.assertTrue(rebar.Shape.isValid())
        self.assertGreater(rebar.Shape.Volume, 0, "Converted rebar shape should have volume.")

    def test_makeRebar_properties_set_correctly(self):
        """Test makeRebar sets dimensional and calculated properties correctly."""
        sketch = self._create_sketch(length=900)
        host = self._create_host_structure(length=1200)

        rebar = Arch.makeRebar(baseobj=host, sketch=sketch, diameter=12.5, amount=7, offset=30.0)
        # Default label "Rebar" will be used, and Mark will also be "Rebar"
        self.document.recompute()

        self.assertEqual(rebar.Mark, "Rebar")
        self.assertAlmostEqual(rebar.Diameter.Value, 12.5, delta=1e-9)
        self.assertEqual(rebar.Amount, 7)
        self.assertAlmostEqual(rebar.OffsetStart.Value, 30.0, delta=1e-9)
        self.assertAlmostEqual(rebar.OffsetEnd.Value, 30.0, delta=1e-9)
        self.assertEqual(rebar.Host, host)
        self.assertEqual(rebar.Base, sketch)

        self.assertTrue(hasattr(rebar, "Length"), "Rebar should have Length property.")
        self.assertTrue(hasattr(rebar, "TotalLength"), "Rebar should have TotalLength property.")
        self.assertAlmostEqual(
            rebar.Length.Value,
            sketch.Shape.Length,
            delta=1e-9,
            msg="Single rebar length should match sketch length.",
        )
        expected_total_length = rebar.Length.Value * rebar.Amount
        self.assertAlmostEqual(
            rebar.TotalLength.Value,
            expected_total_length,
            delta=1e-9,
            msg="TotalLength calculation is incorrect.",
        )

    def test_makeRebar_no_sketch_no_baseobj(self):
        """Test makeRebar with no sketch/baseobj creates object with no meaningful shape."""
        rebar = Arch.makeRebar(diameter=10, amount=1)
        self.document.recompute()

        self.assertIsNotNone(rebar, "makeRebar should create an object even with minimal inputs.")
        self.assertIsNone(rebar.Base, "Rebar.Base should be None.")
        self.assertIsNone(rebar.Host, "Rebar.Host should be None.")
        self.assertAlmostEqual(rebar.Diameter.Value, 10.0, delta=1e-9)
        self.assertEqual(rebar.Amount, 1)

        # A Part::FeaturePython object always has a .Shape attribute.
        # For a rebar with no Base, its execute() returns early, so no geometry is made.
        # The Shape should be the default null/empty shape.
        self.assertIsNotNone(rebar.Shape, "Rebar object should always have a Shape attribute.")
        self.assertTrue(
            rebar.Shape.isNull(), "Rebar.Shape should be a null shape if no Base is provided."
        )

    def test_RebarMe(self):
        """
        Tests the creation of an Arch Rebar
        """
        import Sketcher

        self.printTestMessage("Checking Arch Rebar...\n")
        s = Arch.makeStructure(length=2, width=3, height=5)
        sk = self.document.addObject("Sketcher::SketchObject", "Sketch")
        sk.AttachmentSupport = (s, ["Face6"])
        sk.addGeometry(
            Part.LineSegment(FreeCAD.Vector(-0.85, 1.25, 0), FreeCAD.Vector(0.75, 1.25, 0))
        )
        sk.addGeometry(
            Part.LineSegment(FreeCAD.Vector(0.75, 1.25, 0), FreeCAD.Vector(0.75, -1.20, 0))
        )
        sk.addGeometry(
            Part.LineSegment(FreeCAD.Vector(0.75, -1.20, 0), FreeCAD.Vector(-0.85, -1.20, 0))
        )
        sk.addGeometry(
            Part.LineSegment(FreeCAD.Vector(-0.85, -1.20, 0), FreeCAD.Vector(-0.85, 1.25, 0))
        )
        sk.addConstraint(Sketcher.Constraint("Coincident", 0, 2, 1, 1))
        sk.addConstraint(Sketcher.Constraint("Coincident", 1, 2, 2, 1))
        sk.addConstraint(Sketcher.Constraint("Coincident", 2, 2, 3, 1))
        sk.addConstraint(Sketcher.Constraint("Coincident", 3, 2, 0, 1))
        self.document.recompute()
        r = Arch.makeRebar(s, sk, diameter=0.1, amount=2)
        self.document.recompute()
        self.assertTrue(r, "Arch Rebar creation failed")
        self.assertFalse(r.Shape.isNull(), "Rebar shape is null")



# ***************************************************************************
# *   Copyright (c) 2015 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM constraint self weight document object"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package constraint_selfweight
#  \ingroup FEM
#  \brief constraint self weight object

import FreeCAD
from . import base_fempythonobject
from femtools import constants


class ConstraintSelfWeight(base_fempythonobject.BaseFemPythonObject):
    """
    The ConstraintSelfWeight object"
    """

    Type = "Fem::ConstraintSelfWeight"

    def __init__(self, obj):
        super().__init__(obj)

        self.addProperty(obj)

        # https://wiki.freecad.org/Scripted_objects#Property_Type
        # https://forum.freecad.org/viewtopic.php?f=18&t=13460&start=20#p109709
        # https://forum.freecad.org/viewtopic.php?t=25524
        # obj.setEditorMode("References", 1)  # read only in PropertyEditor, but writeable by Python
        obj.setEditorMode("References", 2)  # do not show in Editor

    def addProperty(self, obj):
        obj.addProperty(
            "App::PropertyAcceleration", "GravityAcceleration", "Gravity", "Gravity acceleration"
        )
        obj.setPropertyStatus("GravityAcceleration", "LockDynamic")
        obj.GravityAcceleration = constants.gravity()

        obj.addProperty(
            "App::PropertyVector", "GravityDirection", "Gravity", "Normalized gravity direction"
        )
        obj.setPropertyStatus("GravityDirection", "LockDynamic")
        obj.GravityDirection = FreeCAD.Vector(0, 0, -1)

        obj.setPropertyStatus("NormalDirection", "Hidden")

    def onDocumentRestored(self, obj):
        # migrate old App::PropertyFloat "Gravity_*" if exists
        try:
            grav_x = obj.getPropertyByName("Gravity_x")
            grav_y = obj.getPropertyByName("Gravity_y")
            grav_z = obj.getPropertyByName("Gravity_z")
            grav = FreeCAD.Vector(grav_x, grav_y, grav_z)

            self.addProperty(obj)
            obj.GravityAcceleration = constants.gravity()
            obj.GravityAcceleration *= grav.Length
            obj.GravityDirection = grav.normalize()

            obj.removeProperty("Gravity_x")
            obj.removeProperty("Gravity_y")
            obj.removeProperty("Gravity_z")

            return

        except:
            return

    def execute(self, obj):
        obj.GravityDirection.normalize()

        return False



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the PathTwistedArray object.

The copies will be placed along a path like a polyline, spline, or bezier
curve, and the copies are twisted around the path by a given rotation
parameter.

This array was developed in order to build a `twisted bridge` object.

See https://forum.freecad.org/viewtopic.php?f=23&t=49617

A `twisted bridge` would consist of three parts:
 1. The ribcage composed of a twisted array generated from a frame
    and a path.
 2. The `tunnel` object produced by lofting or sweeping the internal twisted
    profiles of the ribcage along the path.
 3. The `walkway` object on which the person can stand; it is generated
    from the path, and the internal width of the ribcage profile.

This module builds only the first element, the twisted ribcage.

The tunnel and walkway are built with the `twisted bridge`
object in the Arch Workbench.
"""
## @package pathtwistedarray
# \ingroup draftobjects
# \brief Provides the object code for the TwistedArray object.

import draftgeoutils.geo_arrays as geo
from draftutils.messages import _log


def QT_TRANSLATE_NOOP(ctx, txt):
    return txt


from draftobjects.draftlink import DraftLink

## \addtogroup draftobjects
# @{


class PathTwistedArray(DraftLink):
    """The PathTwistedArray object.

    This array distributes copies of an object along a path like a polyline,
    spline, or bezier curve, and the copies are twisted around the path
    by a given rotation parameter.
    """

    def __init__(self, obj):
        super().__init__(obj, "PathTwistedArray")

    def attach(self, obj):
        """Set up the properties when the object is attached."""
        self.set_properties(obj)
        super().attach(obj)

    def set_properties(self, obj):
        """Set properties only if they don't exist."""
        if hasattr(obj, "PropertiesList"):
            properties = obj.PropertiesList
        else:
            properties = []

        if "Base" not in properties:
            obj.addProperty(
                "App::PropertyLink",
                "Base",
                "Objects",
                QT_TRANSLATE_NOOP("App::Property", "The base object that will be duplicated."),
                locked=True,
            )
            obj.Base = None

        if "PathObject" not in properties:
            obj.addProperty(
                "App::PropertyLink",
                "PathObject",
                "Objects",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The object along which the copies will be distributed. It must contain 'Edges'.",
                ),
                locked=True,
            )
            obj.PathObject = None

        if "Fuse" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Specifies if the copies "
                "should be fused together "
                "if they touch each other (slower)",
            )
            obj.addProperty("App::PropertyBool", "Fuse", "Objects", _tip, locked=True)
            obj.Fuse = False

        if "Count" not in properties:
            obj.addProperty(
                "App::PropertyInteger",
                "Count",
                "Objects",
                QT_TRANSLATE_NOOP("App::Property", "Number of copies to create."),
                locked=True,
            )
            obj.Count = 15

        if "RotationFactor" not in properties:
            obj.addProperty(
                "App::PropertyFloat",
                "RotationFactor",
                "Objects",
                QT_TRANSLATE_NOOP("App::Property", "Rotation factor of the twisted array."),
                locked=True,
            )
            obj.RotationFactor = 0.25

        if self.use_link and "ExpandArray" not in properties:
            obj.addProperty(
                "App::PropertyBool",
                "ExpandArray",
                "Objects",
                QT_TRANSLATE_NOOP(
                    "App::Property", "Show the individual array elements (only for Link arrays)"
                ),
                locked=True,
            )
            obj.ExpandArray = False
            obj.setPropertyStatus("Shape", "Transient")

        if not self.use_link:
            if "PlacementList" not in properties:
                _tip = QT_TRANSLATE_NOOP("App::Property", "The placement for each array element")
                obj.addProperty(
                    "App::PropertyPlacementList", "PlacementList", "Objects", _tip, locked=True
                )
                obj.PlacementList = []

    def linkSetup(self, obj):
        """Set up the object as a link object."""
        super().linkSetup(obj)
        obj.configLinkProperty(ElementCount="Count")

    def onDocumentRestored(self, obj):
        super().onDocumentRestored(obj)
        # Fuse property was added in v1.0 and PlacementList property was added
        # for non-link arrays in v1.1, obj should be OK if both are present:
        if hasattr(obj, "Fuse") and hasattr(obj, "PlacementList"):
            return

        if not hasattr(obj, "Fuse"):
            _log("v1.0, " + obj.Name + ", added 'Fuse' property")
        if not hasattr(obj, "PlacementList"):
            _log("v1.1, " + obj.Name + ", added hidden property 'PlacementList'")

        self.set_properties(obj)
        self.execute(obj)  # Required to update PlacementList.

    def execute(self, obj):
        """Execute when the object is created or recomputed."""
        if self.props_changed_placement_only(obj) or not obj.Base or not obj.PathObject:
            self.props_changed_clear()
            return

        # placement of entire PathArray object
        array_placement = obj.Placement

        path = obj.PathObject
        count = obj.Count
        rot_factor = obj.RotationFactor

        copy_placements, _ = geo.get_twisted_placements(path, count=count, rot_factor=rot_factor)

        self.buildShape(obj, array_placement, copy_placements)
        self.props_changed_clear()
        return not self.use_link


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create Rectangle objects."""
## @package make_rectangle
# \ingroup draftmake
# \brief This module provides the code for Draft make_rectangle function.

## \addtogroup draftmake
# @{
import FreeCAD as App
import draftutils.utils as utils
import draftutils.gui_utils as gui_utils

from draftobjects.rectangle import Rectangle

if App.GuiUp:
    from draftviewproviders.view_rectangle import ViewProviderRectangle


def make_rectangle(length, height=0, placement=None, face=None, support=None):
    """make_rectangle(length, width, [placement], [face])

    Creates a Rectangle object with length in X direction and height in Y
    direction.

    Parameters
    ----------
    length, height : dimensions of the rectangle

    placement : Base.Placement
        If a placement is given, it is used.

    face : Bool
        If face is False, the rectangle is shown as a wireframe,
        otherwise as a face.

    Rectangles can also be constructed by giving them a list of four vertices
    as first argument: make_rectangle(list_of_vertices, face=...)
    but you are responsible to check yourself that these 4 vertices are ordered
    and actually form a rectangle, otherwise the result might be wrong. Placement
    is ignored when constructing a rectangle this way (face argument is kept).
    """

    if not App.ActiveDocument:
        App.Console.PrintError("No active document. Aborting\n")
        return

    if isinstance(length, (list, tuple)) and (len(length) == 4):
        verts = length
        xv = verts[1].sub(verts[0])
        yv = verts[3].sub(verts[0])
        zv = xv.cross(yv)
        rr = App.Rotation(xv, yv, zv, "XYZ")
        rp = App.Placement(verts[0], rr)
        return make_rectangle(xv.Length, yv.Length, rp, face, support)

    if placement:
        utils.type_check([(placement, App.Placement)], "make_rectangle")

    obj = App.ActiveDocument.addObject("Part::Part2DObjectPython", "Rectangle")
    Rectangle(obj)

    obj.Length = length
    obj.Height = height
    obj.AttachmentSupport = support

    if face is not None:
        obj.MakeFace = face

    if placement:
        obj.Placement = placement

    if App.GuiUp:
        ViewProviderRectangle(obj.ViewObject)
        gui_utils.format_object(obj)
        gui_utils.select(obj)

    return obj


makeRectangle = make_rectangle

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# -*- coding: utf8 -*-
## @package importOCA
#  \ingroup DRAFT
#  \brief OCA (Open CAD Format) file importer & exporter
"""@package importOCA
OCA (Open CAD Format) file importer & exporter

This module provides support for importing from and exporting to
the OCA format or GCAD format from GCAD3D (http://www.gcad3d.org/).
See: https://groups.google.com/forum/#!forum/open_cad_format

As of 2019 this file format is practically obsolete, and this module is not
maintained.
"""
# Check code with
# flake8 --ignore=E226,E266,E401,W503

# ***************************************************************************
# *   Copyright (c) 2009 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD Draft Workbench - OCA importer/exporter"
__author__ = "Yorik van Havre <yorik@uncreated.net>"
__url__ = "https://www.freecad.org"

import FreeCAD, os, Part, DraftVecUtils, DraftGeomUtils
from FreeCAD import Vector
from FreeCAD import Console as FCC
from draftutils import params
from draftutils.utils import pyopen

if FreeCAD.GuiUp:
    from draftutils.translate import translate
else:

    def translate(context, txt):
        return txt


def getpoint(data):
    """Turn an OCA point definition into a FreeCAD.Vector.

    Parameters
    ----------
    data : list
        Different types of data.

    Returns
    -------
    Base::Vector3
        A vector with the data arranged, depending on the contents of `data`.
    """
    FCC.PrintMessage("found point %s \n" % data)
    if len(data) == 3:
        return Vector(float(data[0]), float(data[1]), float(data[2]))
    elif (data[0] == "P") and (len(data) == 4):
        return Vector(float(data[1]), float(data[2]), float(data[3]))
    elif (data[0][0] == "P") and (len(data[0]) > 1):
        if len(data) == 1:
            return objects[data[0]]
        else:
            if data[1][0] == "R":
                return objects[data[0]].add(objects[data[1]])
            elif data[1][0] == "C":
                # Error: DraftGeomUtils.findProjection()
                # doesn't exist
                return DraftGeomUtils.findProjection(objects[data[0]], objects[data[1]])
    elif data[0][0] == "C":
        if objects[data[0]]:
            p1 = objects[data[0]].Curve.Position
            if len(data) == 1:
                return p1
            else:
                if data[1][0] == "L":
                    L = objects[data[1]]
                    return p1.add(DraftGeomUtils.vec(L))


def getarea(data):
    """Turn an OCA area definition into a FreeCAD Part.Wire.

    Parameters
    ----------
    data : list
        Different types of data.

    Returns
    -------
    Part.Wire
        A wire object from the points in `data`.
    """
    FCC.PrintMessage("found area %s \n" % data)
    if data[0] == "S":
        if data[1] == "POL":
            pts = data[2:]
            verts = []
            for p in pts:
                if p[0] == "P":
                    verts.append(getpoint([p]))
            w = Part.makePolygon(verts)
            return w


def getarc(data):
    """Turn an OCA arc definition into a FreeCAD Part.Edge.

    Parameters
    ----------
    data : list
        Different types of data.

    Returns
    -------
    Part.Edge
        An edge object from the points in `data`.
    """
    FCC.PrintMessage("found arc %s \n" % data)
    c = None
    if data[0] == "ARC":
        # 3-points arc
        pts = data[1:]
        verts = []
        for p in range(len(pts)):
            if pts[p] == "P":
                verts.append(getpoint(pts[p : p + 3]))
            elif pts[p][0] == "P":
                verts.append(getpoint([pts[p]]))
        if verts[0] and verts[1] and verts[2]:
            c = Part.Arc(verts[0], verts[1], verts[2])
    elif data[0][0] == "P":
        # 2-point circle
        verts = []
        rad = None
        lines = []
        for p in range(len(data)):
            if data[p] == "P":
                verts.append(getpoint(data[p : p + 4]))
            elif data[p][0] == "P":
                verts.append(getpoint([data[p]]))
            elif data[p] == "VAL":
                rad = float(data[p + 1])
            elif data[p][0] == "L":
                lines.append(objects[data[p]])
        c = Part.Circle()
        c.Center = verts[0]
        if rad:
            c.Radius = rad
        else:
            # Error: DraftVecUtils.new()
            # doesn't exist
            c.Radius = DraftVecUtils.new(verts[0], verts[1]).Length
    elif data[0][0] == "L":
        # 2-lines circle
        lines = []
        rad = None
        for p in range(len(data)):
            if data[p] == "VAL":
                rad = float(data[p + 1])
            elif data[p][0] == "L":
                lines.append(objects[data[p]])
        circles = DraftGeomUtils.circleFrom2LinesRadius(lines[0], lines[1], rad)
        if circles:
            c = circles[0]
    if c:
        return c.toShape()


def getline(data):
    """Turns an OCA line definition into a FreeCAD Part.Edge.

    Parameters
    ----------
    data : list
        Different types of data.

    Returns
    -------
    Part.Edge
        An edge object from the points in `data`.
    """
    FCC.PrintMessage("found line %s \n" % data)
    verts = []
    for p in range(len(data)):
        if data[p] == "P":
            verts.append(getpoint(data[p : p + 4]))
        elif data[p][0] == "P":
            verts.append(getpoint([data[p]]))
    L = Part.LineSegment(verts[0], verts[1])
    return L.toShape()


def gettranslation(data):
    """Retrieve a translation (move) vector from `data`.

    Parameters
    ----------
    data : list
        Different types of data.

    Returns
    -------
    Base::Vector3
        A vector with X, Y, or Z displacement, or (0, 0, 0).
    """
    FCC.PrintMessage("found translation %s \n" % data)
    if data[0] == "Z":
        return Vector(0, 0, float(data[1]))
    elif data[0] == "Y":
        return Vector(0, float(data[1]), 0)
    elif data[0] == "X":
        return Vector(float(data[1]), 0, 0)
    return Vector(0, 0, 0)


def writepoint(vector):
    """Write a FreeCAD.Vector in OCA format.

    Parameters
    ----------
    vector : Base::Vector3
        A vector with three components.

    Returns
    -------
    str
        A string "P(X Y Z)" with the information from `vector`.
    """
    return "P(" + str(vector.x) + " " + str(vector.y) + " " + str(vector.z) + ")"


def createobject(oid, doc):
    """Create Part::Feature object in the current document.

    Parameters
    ----------
    oid : str
        ID number of a particular object in the document.

    doc : App::Document
        A FreeCAD document to which the object is added.
    Returns
    -------
    None
    """
    if isinstance(objects[oid], Part.Shape):
        ob = doc.addObject("Part::Feature", oid)
        ob.Shape = objects[oid]
        if FreeCAD.GuiUp:
            ob.ViewObject.ShapeColor = color


def parse(filename, doc):
    """Import an opened OCA file into the given document.

    Parameters
    ----------
    filename : str
        The path to the OCA file.
    doc : App::Document
        A FreeCAD document to which the object is added.

    Returns
    -------
    None
    """
    filebuffer = pyopen(filename)
    global objects
    objects = {}
    global color
    color = (0, 0, 0)
    for l in filebuffer:
        readline = l.replace(",", " ").upper()
        if "=" in readline:
            # entity definitions
            pair = readline.split("=")
            _id = pair[0]
            data = pair[1]
            data = data.replace(",", " ")
            data = data.replace("(", " ")
            data = data.replace(")", " ")
            data = data.split()
            if _id[0] == "P":
                # point
                objects[_id] = getpoint(data)
            elif (_id[0] == "A") and params.get_param("ocaareas"):
                # area
                objects[_id] = getarea(data)
                createobject(_id, doc)

            elif _id[0] == "C":
                # arc or circle
                objects[_id] = getarc(data)
                createobject(_id, doc)

            elif _id[0] == "L":
                # line
                objects[_id] = getline(data)
                createobject(_id, doc)

            elif _id[0] == "R":
                # translation
                objects[_id] = gettranslation(data)

        elif readline[0:6] == "DEFCOL":
            # color
            c = readline.split()
            color = (float(c[1]) / 255, float(c[2]) / 255, float(c[3]) / 255)


def open(filename):
    """Open filename and parse.

    Parameters
    ----------
    filename : str
        The path to the filename to be opened.

    Returns
    -------
    None
    """
    docname = os.path.split(filename)[1]
    doc = FreeCAD.newDocument(docname)
    if docname[-4:] == "gcad":
        doc.Label = docname[:-5]
    else:
        doc.Label = docname[:-4]
    parse(filename, doc)
    doc.recompute()


def insert(filename, docname):
    """Get an active document and parse.

    If no document exist, it is created.

    Parameters
    ----------
    filename : str
        The path to the filename to be opened.
    docname : str
        The name of the active App::Document if one exists, or
        of the new one created.

    Returns
    -------
    None
    """
    try:
        doc = FreeCAD.getDocument(docname)
    except NameError:
        doc = FreeCAD.newDocument(docname)
    FreeCAD.ActiveDocument = doc
    parse(filename, doc)
    doc.recompute()


def export(exportList, filename):
    """Export the OCA file with a given list of objects.

    The objects must be edges or faces, in order to be processed
    and exported.

    Parameters
    ----------
    exportList : list
        List of document objects to export.
    filename : str
        Path to the new file.

    Returns
    -------
    None
        If `exportList` doesn't have shapes to export.
    """
    faces = []
    edges = []

    # getting faces and edges
    for ob in exportList:
        if ob.Shape.Faces:
            for f in ob.Shape.Faces:
                faces.append(f)
        else:
            for e in ob.Shape.Edges:
                edges.append(e)
    if not (edges or faces):
        FCC.PrintMessage(translate("importOCA", "OCA: found no data to export") + "\n")
        return

    # writing file
    oca = pyopen(filename, "w")
    oca.write("#oca file generated from FreeCAD\r\n")
    oca.write("# edges\r\n")
    count = 1
    for e in edges:
        if DraftGeomUtils.geomType(e) == "Line":
            oca.write("L" + str(count) + "=")
            oca.write(writepoint(e.Vertexes[0].Point))
            oca.write(" ")
            oca.write(writepoint(e.Vertexes[-1].Point))
            oca.write("\r\n")
        elif DraftGeomUtils.geomType(e) == "Circle":
            if len(e.Vertexes) > 1:
                oca.write("C" + str(count) + "=ARC ")
                oca.write(writepoint(e.Vertexes[0].Point))
                oca.write(" ")
                oca.write(writepoint(DraftGeomUtils.findMidpoint(e)))
                oca.write(" ")
                oca.write(writepoint(e.Vertexes[-1].Point))
            else:
                oca.write("C" + str(count) + "= ")
                oca.write(writepoint(e.Curve.Center))
                oca.write(" ")
                oca.write(str(e.Curve.Radius))
            oca.write("\r\n")
        count += 1
    oca.write("# faces\r\n")
    for f in faces:
        oca.write("A" + str(count) + "=S(POL")
        for v in f.Vertexes:
            oca.write(" ")
            oca.write(writepoint(v.Point))
        oca.write(" ")
        oca.write(writepoint(f.Vertexes[0].Point))
        oca.write(")\r\n")
        count += 1

    # closing
    oca.close()
    FCC.PrintMessage(translate("importOCA", "successfully exported") + " " + filename + "\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2016 sliptonic <shopinthewoods@gmail.com>               *
# *   Copyright (c) 2021 Schildkroet                                        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import math

from FreeCAD import Vector
from PySide import QtCore

# lazily loaded modules
from lazy_loader.lazy_loader import LazyLoader

Part = LazyLoader("Part", globals(), "Part")

__title__ = "Geom - geometry utilities for CAM"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Functions to extract and convert between Path.Command and Part.Edge and utility functions to reason about them."

Tolerance = 0.000001

translate = FreeCAD.Qt.translate

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class Side:
    """Class to determine and define the side a Path is on, or Vectors are in relation to each other."""

    Left = +1
    Right = -1
    Straight = 0
    On = 0

    @classmethod
    def toString(cls, side):
        """toString(side)
        Returns a string representation of the enum value."""
        if side == cls.Left:
            return "Left"
        if side == cls.Right:
            return "Right"
        return "On"

    @classmethod
    def of(cls, ptRef, pt):
        """of(ptRef, pt)
        Determine the side of pt in relation to ptRef.
        If both Points are viewed as vectors with their origin in (0,0,0)
        then the two vectors either form a straight line (On) or pt
        lies in the left or right hemisphere in regards to ptRef."""
        d = -ptRef.x * pt.y + ptRef.y * pt.x
        if d < 0:
            return cls.Left
        if d > 0:
            return cls.Right
        return cls.Straight


CmdMoveRapid = ["G0", "G00"]
CmdMoveStraight = ["G1", "G01"]
CmdMoveCW = ["G2", "G02"]
CmdMoveCCW = ["G3", "G03"]
CmdMoveDrill = ["G73", "G81", "G82", "G83", "G85"]
CmdMoveArc = CmdMoveCW + CmdMoveCCW
CmdMove = CmdMoveStraight + CmdMoveArc + CmdMoveDrill
CmdMoveAll = CmdMove + CmdMoveRapid


def isRoughly(float1, float2, error=Tolerance):
    """isRoughly(float1, float2, [error=Tolerance])
    Returns true if the two values are the same within a given error."""
    return math.fabs(float1 - float2) <= error


def pointsCoincide(p1, p2, error=Tolerance):
    """pointsCoincide(p1, p2, [error=Tolerance])
    Return True if two points are roughly identical (see also isRoughly)."""
    return (
        isRoughly(p1.x, p2.x, error)
        and isRoughly(p1.y, p2.y, error)
        and isRoughly(p1.z, p2.z, error)
    )


def edgesMatch(e0, e1, error=Tolerance):
    """edgesMatch(e0, e1, [error=Tolerance]
    Return true if the edges start and end at the same point and have the same type of curve."""
    if type(e0.Curve) != type(e1.Curve) or len(e0.Vertexes) != len(e1.Vertexes):
        return False
    return all(
        pointsCoincide(e0.Vertexes[i].Point, e1.Vertexes[i].Point, error)
        for i in range(len(e0.Vertexes))
    )


def edgeConnectsTo(edge, vector, error=Tolerance):
    """edgeConnectsTop(edge, vector, error=Tolerance)
    Returns True if edge connects to given vector."""
    return pointsCoincide(edge.valueAt(edge.FirstParameter), vector, error) or pointsCoincide(
        edge.valueAt(edge.LastParameter), vector, error
    )


def normalizeAngle(a):
    """normalizeAngle(a) ... return angle shifted into interval -pi <= a <= pi"""
    while a > math.pi:
        a = a - 2 * math.pi
    while a < -math.pi:
        a = a + 2 * math.pi
    return a


def getAngle(vector):
    """getAngle(vector)
    Returns the angle [-pi,pi] of a vector using the X-axis as the reference.
    Positive angles for vertexes in the upper hemisphere (positive y values)
    and negative angles for the lower hemisphere."""
    a = vector.getAngle(Vector(1, 0, 0))
    if vector.y < 0:
        return -a
    return a


def diffAngle(a1, a2, direction="CW"):
    """diffAngle(a1, a2, [direction='CW'])
    Returns the difference between two angles (a1 -> a2) into a given direction."""
    if direction == "CW":
        while a1 < a2:
            a1 += 2 * math.pi
        a = a1 - a2
    else:
        while a2 < a1:
            a2 += 2 * math.pi
        a = a2 - a1
    return a


def isVertical(obj):
    """isVertical(obj) ... answer True if obj points into Z"""
    if type(obj) == FreeCAD.Vector:
        return isRoughly(obj.x, 0) and isRoughly(obj.y, 0)

    if obj.ShapeType == "Face":
        if type(obj.Surface) == Part.Plane:
            return isHorizontal(obj.Surface.Axis)
        if type(obj.Surface) == Part.Cylinder or type(obj.Surface) == Part.Cone:
            return isVertical(obj.Surface.Axis)
        if type(obj.Surface) == Part.Sphere:
            return True
        if type(obj.Surface) == Part.SurfaceOfExtrusion:
            return isVertical(obj.Surface.Direction)
        if type(obj.Surface) == Part.SurfaceOfRevolution:
            return isHorizontal(obj.Surface.Direction)
        if type(obj.Surface) != Part.BSplineSurface:
            Path.Log.info(
                translate("PathGeom", "face %s not handled, assuming not vertical")
                % type(obj.Surface)
            )
        return None

    if obj.ShapeType == "Edge":
        if type(obj.Curve) == Part.Line or type(obj.Curve) == Part.LineSegment:
            return isVertical(obj.Vertexes[1].Point - obj.Vertexes[0].Point)
        if (
            type(obj.Curve) == Part.Circle or type(obj.Curve) == Part.Ellipse
        ):  # or type(obj.Curve) == Part.BSplineCurve:
            return isHorizontal(obj.Curve.Axis)
        if type(obj.Curve) == Part.BezierCurve:
            # the current assumption is that a bezier curve is vertical if its end points are vertical
            return isVertical(obj.Curve.EndPoint - obj.Curve.StartPoint)
        if type(obj.Curve) != Part.BSplineCurve:
            Path.Log.info(
                translate("PathGeom", "edge %s not handled, assuming not vertical")
                % type(obj.Curve)
            )
        return None

    Path.Log.error(translate("PathGeom", "isVertical(%s) not supported") % obj)
    return None


def isHorizontal(obj):
    """isHorizontal(obj) ... answer True if obj points into X or Y"""
    if type(obj) == FreeCAD.Vector:
        return isRoughly(obj.z, 0)

    if obj.ShapeType == "Face":
        if type(obj.Surface) == Part.Plane:
            return isVertical(obj.Surface.Axis)
        if type(obj.Surface) == Part.Cylinder or type(obj.Surface) == Part.Cone:
            return isHorizontal(obj.Surface.Axis)
        if type(obj.Surface) == Part.Sphere:
            return True
        if type(obj.Surface) == Part.SurfaceOfExtrusion:
            return isHorizontal(obj.Surface.Direction)
        if type(obj.Surface) == Part.SurfaceOfRevolution:
            return isVertical(obj.Surface.Direction)
        return isRoughly(obj.BoundBox.ZLength, 0.0)

    if obj.ShapeType == "Edge":
        if type(obj.Curve) == Part.Line or type(obj.Curve) == Part.LineSegment:
            return isHorizontal(obj.Vertexes[1].Point - obj.Vertexes[0].Point)
        if (
            type(obj.Curve) == Part.Circle or type(obj.Curve) == Part.Ellipse
        ):  # or type(obj.Curve) == Part.BSplineCurve:
            return isVertical(obj.Curve.Axis)
        return isRoughly(obj.BoundBox.ZLength, 0.0)

    Path.Log.error(translate("PathGeom", "isHorizontal(%s) not supported") % obj)
    return None


def commandEndPoint(cmd, defaultPoint=Vector(), X="X", Y="Y", Z="Z"):
    """commandEndPoint(cmd, [defaultPoint=Vector()], [X='X'], [Y='Y'], [Z='Z'])
    Extracts the end point from a Path Command."""
    x = cmd.Parameters.get(X, defaultPoint.x)
    y = cmd.Parameters.get(Y, defaultPoint.y)
    z = cmd.Parameters.get(Z, defaultPoint.z)
    return Vector(x, y, z)


def xy(point):
    """xy(point)
    Convenience function to return the projection of the Vector in the XY-plane."""
    return Vector(point.x, point.y, 0)


def speedBetweenPoints(p0, p1, hSpeed, vSpeed):
    if isRoughly(hSpeed, vSpeed):
        return hSpeed

    d = p1 - p0
    if isRoughly(0.0, d.z):
        return hSpeed
    if isRoughly(0.0, d.x) and isRoughly(0.0, d.y):
        return vSpeed
    # need to interpolate between hSpeed and vSpeed depending on the pitch
    pitch = 2 * math.atan2(xy(d).Length, math.fabs(d.z)) / math.pi
    while pitch < 0:
        pitch = pitch + 1
    while pitch > 1:
        pitch = pitch - 1
    Path.Log.debug(
        "  pitch = %g %g (%.2f, %.2f, %.2f) -> %.2f"
        % (pitch, math.atan2(xy(d).Length, d.z), d.x, d.y, d.z, xy(d).Length)
    )
    speed = vSpeed + pitch * (hSpeed - vSpeed)
    if speed > hSpeed and speed > vSpeed:
        return max(hSpeed, vSpeed)
    if speed < hSpeed and speed < vSpeed:
        return min(hSpeed, vSpeed)
    return speed


def cmdsForEdge(edge, flip=False, useHelixForBSpline=True, segm=50, hSpeed=0, vSpeed=0):
    """cmdsForEdge(edge, flip=False, useHelixForBSpline=True, segm=50) -> List(Path.Command)
    Returns a list of Path.Command representing the given edge.
    If flip is True the edge is considered to be backwards.
    If useHelixForBSpline is True an Edge based on a BSplineCurve is considered
    to represent a helix and results in G2 or G3 command. Otherwise edge has
    no direct Path.Command mapping and will be approximated by straight segments.
    segm is a factor for the segmentation of arbitrary curves not mapped to G1/2/3
    commands. The higher the value the more segments will be used."""
    pt = edge.valueAt(edge.LastParameter) if not flip else edge.valueAt(edge.FirstParameter)
    params = {"X": pt.x, "Y": pt.y, "Z": pt.z}
    if type(edge.Curve) == Part.Line or type(edge.Curve) == Part.LineSegment:
        if hSpeed > 0 and vSpeed > 0:
            pt2 = (
                edge.valueAt(edge.FirstParameter) if not flip else edge.valueAt(edge.LastParameter)
            )
            params.update({"F": speedBetweenPoints(pt, pt2, hSpeed, vSpeed)})
        commands = [Path.Command("G1", params)]
    else:
        p1 = edge.valueAt(edge.FirstParameter) if not flip else edge.valueAt(edge.LastParameter)
        p2 = edge.valueAt((edge.FirstParameter + edge.LastParameter) / 2)
        p3 = pt

        if hasattr(edge.Curve, "Axis") and (
            (
                type(edge.Curve) == Part.Circle
                and isRoughly(edge.Curve.Axis.x, 0)
                and isRoughly(edge.Curve.Axis.y, 0)
            )
            or (useHelixForBSpline and type(edge.Curve) == Part.BSplineCurve)
        ):
            # This is an arc or a helix and it should be represented by a simple G2/G3 command
            if edge.Curve.Axis.z < 0:
                cmd = "G2" if not flip else "G3"
            else:
                cmd = "G3" if not flip else "G2"

            if pointsCoincide(p1, p3):
                # A full circle
                offset = edge.Curve.Center - pt
            else:
                pd = Part.Circle(xy(p1), xy(p2), xy(p3)).Center
                # Path.Log.debug(
                #    "**** %s.%d: (%.2f, %.2f, %.2f) - (%.2f, %.2f, %.2f) - (%.2f, %.2f, %.2f) -> center=(%.2f, %.2f)"
                #    % (
                #        cmd,
                #        flip,
                #        p1.x,
                #        p1.y,
                #        p1.z,
                #        p2.x,
                #        p2.y,
                #        p2.z,
                #        p3.x,
                #        p3.y,
                #        p3.z,
                #        pd.x,
                #        pd.y,
                #    )
                # )

                # Have to calculate the center in the XY plane, using pd leads to an error if this is a helix
                pa = xy(p1)
                pb = xy(p2)
                pc = xy(p3)
                offset = Part.Circle(pa, pb, pc).Center - pa

                # Path.Log.debug(
                #    "**** (%.2f, %.2f, %.2f) - (%.2f, %.2f, %.2f)"
                #    % (pa.x, pa.y, pa.z, pc.x, pc.y, pc.z)
                # )
                # Path.Log.debug(
                #    "**** (%.2f, %.2f, %.2f) - (%.2f, %.2f, %.2f)"
                #    % (pb.x, pb.y, pb.z, pd.x, pd.y, pd.z)
                # )
            # Path.Log.debug("**** (%.2f, %.2f, %.2f)" % (offset.x, offset.y, offset.z))

            params.update({"I": offset.x, "J": offset.y, "K": (p3.z - p1.z) / 2})
            # G2/G3 commands are always performed at hSpeed
            if hSpeed > 0:
                params.update({"F": hSpeed})
            commands = [Path.Command(cmd, params)]

        else:
            # We're dealing with a helix or a more complex shape and it has to get approximated
            # by a number of straight segments
            points = edge.discretize(Deflection=0.01)
            if flip:
                points = points[::-1]

            commands = []
            if points:
                p0 = points[0]
                for p in points[1:]:
                    params = {"X": p.x, "Y": p.y, "Z": p.z}
                    if hSpeed > 0 and vSpeed > 0:
                        params["F"] = speedBetweenPoints(p0, p, hSpeed, vSpeed)
                    cmd = Path.Command("G1", params)
                    # print("***** {}".format(cmd))
                    commands.append(cmd)
                    p0 = p
    # print commands
    return commands


def edgeForCmd(cmd, startPoint):
    """edgeForCmd(cmd, startPoint).
    Returns an Edge representing the given command, assuming a given startPoint."""

    # Path.Log.debug("cmd: {}".format(cmd))
    # Path.Log.debug("startpoint {}".format(startPoint))

    endPoint = commandEndPoint(cmd, startPoint)
    if (cmd.Name in CmdMoveStraight) or (cmd.Name in CmdMoveRapid) or (cmd.Name in CmdMoveDrill):
        if pointsCoincide(startPoint, endPoint):
            return None
        return Part.Edge(Part.LineSegment(startPoint, endPoint))

    if cmd.Name in CmdMoveArc:
        center = startPoint + commandEndPoint(cmd, Vector(0, 0, 0), "I", "J", "K")
        A = xy(startPoint - center)
        B = xy(endPoint - center)
        d = -B.x * A.y + B.y * A.x

        if isRoughly(d, 0, 0.005):
            # Path.Log.debug(
            #    "Half circle arc at: (%.2f, %.2f, %.2f)" % (center.x, center.y, center.z)
            # )
            # we're dealing with half a circle here
            angle = getAngle(A) + math.pi / 2
            if cmd.Name in CmdMoveCW:
                angle -= math.pi
        else:
            C = A + B
            angle = getAngle(C)
            # Path.Log.debug(
            #    "Arc (%8f) at: (%.2f, %.2f, %.2f) -> angle=%f"
            #    % (d, center.x, center.y, center.z, angle / math.pi)
            # )

        R = A.Length
        # Path.Log.debug(
        #    "arc: p1=(%.2f, %.2f) p2=(%.2f, %.2f) -> center=(%.2f, %.2f)"
        #    % (startPoint.x, startPoint.y, endPoint.x, endPoint.y, center.x, center.y)
        # )
        # Path.Log.debug("arc: A=(%.2f, %.2f) B=(%.2f, %.2f) -> d=%.2f" % (A.x, A.y, B.x, B.y, d))
        # Path.Log.debug("arc: R=%.2f angle=%.2f" % (R, angle / math.pi))
        if isRoughly(startPoint.z, endPoint.z):
            midPoint = center + Vector(math.cos(angle), math.sin(angle), 0) * R
            # Path.Log.debug(
            #    "arc: (%.2f, %.2f) -> (%.2f, %.2f) -> (%.2f, %.2f)"
            #    % (
            #        startPoint.x,
            #        startPoint.y,
            #        midPoint.x,
            #        midPoint.y,
            #        endPoint.x,
            #        endPoint.y,
            #    )
            # )
            # Path.Log.debug("StartPoint:{}".format(startPoint))
            # Path.Log.debug("MidPoint:{}".format(midPoint))
            # Path.Log.debug("EndPoint:{}".format(endPoint))

            if pointsCoincide(startPoint, endPoint, 0.001):
                return Part.makeCircle(R, center, FreeCAD.Vector(0, 0, 1))
            else:
                return Part.Edge(Part.Arc(startPoint, midPoint, endPoint))

        # It's a Helix
        # print('angle: A=%.2f B=%.2f' % (getAngle(A)/math.pi, getAngle(B)/math.pi))
        if cmd.Name in CmdMoveCW:
            cw = True
        else:
            cw = False
        angle = diffAngle(getAngle(A), getAngle(B), "CW" if cw else "CCW")
        height = endPoint.z - startPoint.z
        pitch = height * math.fabs(2 * math.pi / angle)
        if angle > 0:
            cw = not cw
        # print("Helix: R=%.2f h=%.2f angle=%.2f pitch=%.2f" % (R, height, angle/math.pi, pitch))
        helix = Part.makeHelix(pitch, height, R, 0, not cw)
        helix.rotate(Vector(), Vector(0, 0, 1), 180 * getAngle(A) / math.pi)
        e = helix.Edges[0]
        helix.translate(startPoint - e.valueAt(e.FirstParameter))
        return helix.Edges[0]
    return None


def wireForPath(path, startPoint=Vector(0, 0, 0)):
    """wireForPath(path, [startPoint=Vector(0,0,0)])
    Returns a wire representing all move commands found in the given path."""
    edges = []
    rapid = []
    rapid_indexes = set()
    if hasattr(path, "Commands"):
        for cmd in path.Commands:
            edge = edgeForCmd(cmd, startPoint)
            if edge:
                if cmd.Name in CmdMoveRapid:
                    rapid.append(edge)
                    rapid_indexes.add(len(edges))
                edges.append(edge)
                startPoint = commandEndPoint(cmd, startPoint)
    if not edges:
        return (None, rapid, rapid_indexes)
    return (Part.Wire(edges), rapid, rapid_indexes)


def wiresForPath(path, startPoint=Vector(0, 0, 0)):
    """wiresForPath(path, [startPoint=Vector(0,0,0)])
    Returns a collection of wires, each representing a continuous cutting Path in path."""
    wires = []
    if hasattr(path, "Commands"):
        edges = []
        for cmd in path.Commands:
            if cmd.Name in CmdMove:
                edges.append(edgeForCmd(cmd, startPoint))
                startPoint = commandEndPoint(cmd, startPoint)
            elif cmd.Name in CmdMoveRapid:
                if len(edges) > 0:
                    wires.append(Part.Wire(edges))
                    edges = []
                startPoint = commandEndPoint(cmd, startPoint)
        if edges:
            wires.append(Part.Wire(edges))
    return wires


def arcToHelix(edge, z0, z1):
    """arcToHelix(edge, z0, z1)
    Assuming edge is an arc it'll return a helix matching the arc starting at z0 and rising/falling to z1.
    """

    p1 = edge.valueAt(edge.FirstParameter)
    # p2 = edge.valueAt(edge.LastParameter)

    cmd = cmdsForEdge(edge)[0]
    params = cmd.Parameters
    params.update({"Z": z1, "K": (z1 - z0) / 2})
    command = Path.Command(cmd.Name, params)

    # print("- (%.2f, %.2f, %.2f) - (%.2f, %.2f, %.2f): %.2f:%.2f" % (edge.Vertexes[0].X, edge.Vertexes[0].Y, edge.Vertexes[0].Z, edge.Vertexes[1].X, edge.Vertexes[1].Y, edge.Vertexes[1].Z, z0, z1))
    # print("- %s -> %s" % (cmd, command))

    return edgeForCmd(command, Vector(p1.x, p1.y, z0))


def helixToArc(edge, z=0):
    """helixToArc(edge, z=0)
    Returns the projection of the helix onto the XY-plane with a given offset."""
    p1 = edge.valueAt(edge.FirstParameter)
    p2 = edge.valueAt((edge.FirstParameter + edge.LastParameter) / 2)
    p3 = edge.valueAt(edge.LastParameter)
    p01 = Vector(p1.x, p1.y, z)
    p02 = Vector(p2.x, p2.y, z)
    p03 = Vector(p3.x, p3.y, z)
    return Part.Edge(Part.Arc(p01, p02, p03))


def splitArcAt(edge, pt):
    """splitArcAt(edge, pt)
    Returns a list of 2 edges which together form the original arc split at the given point.
    The Vector pt has to represent a point on the given arc."""
    p = edge.Curve.parameter(pt)
    e0 = Part.Arc(edge.Curve.copy(), edge.FirstParameter, p).toShape()
    e1 = Part.Arc(edge.Curve.copy(), p, edge.LastParameter).toShape()
    return [e0, e1]


def splitEdgeAt(edge, pt):
    """splitEdgeAt(edge, pt)
    Returns a list of 2 edges, forming the original edge split at the given point.
    The results are undefined if the Vector representing the point is not part of the edge."""
    # I could not get the OCC parameterAt and split to work ...
    # pt HAS to be on the edge, otherwise the results are undefined
    p1 = edge.valueAt(edge.FirstParameter)
    p2 = pt
    p3 = edge.valueAt(edge.LastParameter)
    # edges = []

    if type(edge.Curve) == Part.Line or type(edge.Curve) == Part.LineSegment:
        # it's a line
        return [
            Part.Edge(Part.LineSegment(p1, p2)),
            Part.Edge(Part.LineSegment(p2, p3)),
        ]
    elif type(edge.Curve) == Part.Circle:
        # it's an arc
        return splitArcAt(edge, pt)
    else:
        # it's a helix
        arc = helixToArc(edge, 0)
        aes = splitArcAt(arc, Vector(pt.x, pt.y, 0))
        return [arcToHelix(aes[0], p1.z, p2.z), arcToHelix(aes[1], p2.z, p3.z)]


def combineConnectedShapes(shapes):
    done = False
    while not done:
        done = True
        combined = []
        Path.Log.debug("shapes: {}".format(shapes))
        for shape in shapes:
            connected = [f for f in combined if isRoughly(shape.distToShape(f)[0], 0.0)]
            Path.Log.debug(
                "  {}: connected: {} dist: {}".format(
                    len(combined),
                    connected,
                    [shape.distToShape(f)[0] for f in combined],
                )
            )
            if connected:
                combined = [f for f in combined if f not in connected]
                connected.append(shape)
                combined.append(Part.makeCompound(connected))
                done = False
            else:
                combined.append(shape)
        shapes = combined
    return shapes


def removeDuplicateEdges(wire):
    unique = []
    for e in wire.Edges:
        if not any(edgesMatch(e, u) for u in unique):
            unique.append(e)
    return Part.Wire(unique)


def flipEdge(edge):
    """flipEdge(edge)
    Flips given edge around so the new Vertexes[0] was the old Vertexes[-1] and vice versa, without changing the shape.
    Currently only lines, line segments, circles and arcs are supported."""

    if Part.Line == type(edge.Curve) and not edge.Vertexes:
        return Part.Edge(
            Part.Line(edge.valueAt(edge.LastParameter), edge.valueAt(edge.FirstParameter))
        )
    elif Part.Line == type(edge.Curve) or Part.LineSegment == type(edge.Curve):
        return Part.Edge(Part.LineSegment(edge.Vertexes[-1].Point, edge.Vertexes[0].Point))
    elif Part.Circle == type(edge.Curve):
        # Create an inverted circle
        circle = Part.Circle(edge.Curve.Center, -edge.Curve.Axis, edge.Curve.Radius)
        # Rotate the circle appropriately so it starts at edge.valueAt(edge.LastParameter)
        circle.rotate(
            FreeCAD.Placement(
                circle.Center,
                circle.Axis,
                180 - math.degrees(edge.LastParameter + edge.Curve.AngleXU),
            )
        )
        # Now the edge always starts at 0 and LastParameter is the value range
        arc = Part.Edge(circle, 0, edge.LastParameter - edge.FirstParameter)
        return arc
    elif type(edge.Curve) in [Part.BSplineCurve, Part.BezierCurve]:
        if type(edge.Curve) == Part.BSplineCurve:
            spline = edge.Curve
        else:
            spline = edge.Curve.toBSpline()

        mults = spline.getMultiplicities()
        weights = spline.getWeights()
        knots = spline.getKnots()
        poles = spline.getPoles()
        perio = spline.isPeriodic()
        ratio = spline.isRational()
        degree = spline.Degree

        ma = max(knots)
        mi = min(knots)
        knots = [ma + mi - k for k in knots]

        mults.reverse()
        weights.reverse()
        poles.reverse()
        knots.reverse()

        flipped = Part.BSplineCurve()
        flipped.buildFromPolesMultsKnots(poles, mults, knots, perio, degree, weights, ratio)

        return Part.Edge(flipped, ma + mi - edge.LastParameter, ma + mi - edge.FirstParameter)
    elif type(edge.Curve) == Part.OffsetCurve:
        return edge.reversed()

    Path.Log.warning(translate("PathGeom", "%s not supported for flipping") % type(edge.Curve))


def flipWire(wire):
    """Flip the entire wire and all its edges so it is being processed the other way around."""
    edges = [flipEdge(e) for e in wire.Edges]
    edges.reverse()
    Path.Log.debug(edges)
    return Part.Wire(edges)


def makeBoundBoxFace(bBox, offset=0.0, zHeight=0.0):
    """makeBoundBoxFace(bBox, offset=0.0, zHeight=0.0)...
    Function to create boundbox face, with possible extra offset and custom Z-height."""
    p1 = FreeCAD.Vector(bBox.XMin - offset, bBox.YMin - offset, zHeight)
    p2 = FreeCAD.Vector(bBox.XMax + offset, bBox.YMin - offset, zHeight)
    p3 = FreeCAD.Vector(bBox.XMax + offset, bBox.YMax + offset, zHeight)
    p4 = FreeCAD.Vector(bBox.XMin - offset, bBox.YMax + offset, zHeight)

    L1 = Part.makeLine(p1, p2)
    L2 = Part.makeLine(p2, p3)
    L3 = Part.makeLine(p3, p4)
    L4 = Part.makeLine(p4, p1)

    return Part.Face(Part.Wire([L1, L2, L3, L4]))


# Method to combine faces if connected
def combineHorizontalFaces(faces):
    """combineHorizontalFaces(faces)...
    This function successfully identifies and combines multiple connected faces and
    works on multiple independent faces with multiple connected faces within the list.
    The return value is a list of simplified faces.
    The Adaptive op is not concerned with which hole edges belong to which face.

    Attempts to do the same shape connecting failed with TechDraw.findShapeOutline() and
    Path.Geom.combineConnectedShapes(), so this algorithm was created.
    """
    horizontal = list()
    offset = 10.0
    topFace = None
    innerFaces = list()

    # Verify all incoming faces are at Z=0.0
    for f in faces:
        if f.BoundBox.ZMin != 0.0:
            f.translate(FreeCAD.Vector(0.0, 0.0, 0.0 - f.BoundBox.ZMin))

    # Make offset compound boundbox solid and cut incoming face extrusions from it
    allFaces = Part.makeCompound(faces)
    if hasattr(allFaces, "Area") and isRoughly(allFaces.Area, 0.0):
        msg = translate(
            "PathGeom",
            "Zero working area to process. Check your selection and settings.",
        )
        Path.Log.info(msg)
        return horizontal

    afbb = allFaces.BoundBox
    bboxFace = makeBoundBoxFace(afbb, offset, -5.0)
    bboxSolid = bboxFace.extrude(FreeCAD.Vector(0.0, 0.0, 10.0))
    extrudedFaces = list()
    for f in faces:
        extrudedFaces.append(f.extrude(FreeCAD.Vector(0.0, 0.0, 6.0)))

    # Fuse all extruded faces together
    allFacesSolid = extrudedFaces.pop()
    for i in range(len(extrudedFaces)):
        temp = extrudedFaces.pop().fuse(allFacesSolid)
        allFacesSolid = temp
    cut = bboxSolid.cut(allFacesSolid)

    # Debug
    # Part.show(cut)
    # FreeCAD.ActiveDocument.ActiveObject.Label = "cut"

    # Identify top face and floating inner faces that are the holes in incoming faces
    for f in cut.Faces:
        fbb = f.BoundBox
        if isRoughly(fbb.ZMin, 5.0) and isRoughly(fbb.ZMax, 5.0):
            if (
                isRoughly(afbb.XMin - offset, fbb.XMin)
                and isRoughly(afbb.XMax + offset, fbb.XMax)
                and isRoughly(afbb.YMin - offset, fbb.YMin)
                and isRoughly(afbb.YMax + offset, fbb.YMax)
            ):
                topFace = f
            else:
                innerFaces.append(f)

    if not topFace:
        return horizontal

    outer = [Part.Face(w) for w in topFace.Wires[1:]]

    if outer:
        for f in outer:
            f.translate(FreeCAD.Vector(0.0, 0.0, 0.0 - f.BoundBox.ZMin))

        if innerFaces:
            # inner = [Part.Face(f.Wire1) for f in innerFaces]
            inner = innerFaces

            for f in inner:
                f.translate(FreeCAD.Vector(0.0, 0.0, 0.0 - f.BoundBox.ZMin))
            innerComp = Part.makeCompound(inner)
            outerComp = Part.makeCompound(outer)
            cut = outerComp.cut(innerComp)
            for f in cut.Faces:
                horizontal.append(f)
        else:
            horizontal = outer

    return horizontal



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import Draft
import FreeCAD
import FreeCADGui
import Part
import Path
import Path.Op.Base as OpBase
import PathScripts.PathUtils as PathUtils

from PySide.QtCore import QT_TRANSLATE_NOOP

__title__ = "CAM Path from Shape with Tool Controller"
__author__ = ""
__inspirer__ = "Russ4262"
__url__ = "https://forum.freecad.org/viewtopic.php?t=93896"
__doc__ = ""


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


translate = FreeCAD.Qt.translate


# Add base set of operation properties
def _addBaseProperties(obj):
    obj.addProperty(
        "App::PropertyBool",
        "Active",
        "Path",
        QT_TRANSLATE_NOOP("App::Property", "Make False, to prevent operation from generating code"),
        locked=True,
    )
    obj.addProperty(
        "App::PropertyString",
        "Comment",
        "Path",
        QT_TRANSLATE_NOOP("App::Property", "An optional comment for this operation"),
        locked=True,
    )
    obj.addProperty(
        "App::PropertyString",
        "UserLabel",
        "Path",
        QT_TRANSLATE_NOOP("App::Property", "User assigned label"),
        locked=True,
    )
    obj.addProperty(
        "App::PropertyString",
        "CycleTime",
        "Path",
        QT_TRANSLATE_NOOP("App::Property", "Operations cycle time estimation"),
        locked=True,
    )
    obj.setEditorMode("CycleTime", 1)  # Set property read-only
    obj.Active = True


# Add ToolController properties
def _addToolController(obj):
    obj.addProperty(
        "App::PropertyLink",
        "ToolController",
        "Path",
        QT_TRANSLATE_NOOP(
            "App::Property",
            "The tool controller that will be used to calculate the path",
        ),
    )
    obj.addProperty(
        "App::PropertyDistance",
        "OpToolDiameter",
        "Op Values",
        QT_TRANSLATE_NOOP("App::Property", "Holds the diameter of the tool"),
    )
    obj.setEditorMode("OpToolDiameter", 1)  # Set property read-only
    obj.ToolController = PathUtils.findToolController(obj, None)
    if not obj.ToolController:
        raise OpBase.PathNoTCException()
    obj.OpToolDiameter = obj.ToolController.Tool.Diameter

    obj.FeedRate = obj.ToolController.HorizFeed.Value
    obj.FeedRateVertical = obj.ToolController.VertFeed.Value


# Get list of tool controllers
def _getToolControllers(obj, proxy=None):
    # Modified getToolControllers() from PathScripts.PathUtils
    # for Path object without Proxy
    job = PathUtils.findParentJob(obj)
    if job:
        return [tc for tc in job.Tools.Group]
    else:
        return []


# Set safety height parameters for Path operation
def _setSafetyZ(obj):
    job = PathUtils.findParentJob(obj)
    if job:
        safetyZ = job.Stock.Shape.BoundBox.ZMax + 10
        obj.RetractThreshold = safetyZ
        obj.Retraction = safetyZ
        obj.ResumeHeight = safetyZ


# Geometry for selected shapes
class ObjectPartShape:
    def __init__(self, obj, base):
        # Path.Log.info("ObjectPartShape.__init__()")
        self.obj = obj
        obj.addProperty(
            "App::PropertyLinkSubListGlobal",
            "Base",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "The base geometry for this operation"),
        )
        obj.Base = base

    def __getstate__(self):
        return None

    def __setstate__(self, state):
        return None

    def onDelete(self, obj, args):
        return True

    def onDocumentRestored(self, obj):
        self.obj = obj

    def onChanged(self, obj, prop):
        """onChanged(obj, prop) ... method called when objECT is changed,
        with source propERTY of the change."""
        if "Restore" in obj.State:
            pass

    def execute(self, obj):
        edges = []
        if obj.Base:
            (base, subNames) = obj.Base[0]
            edges = [
                base.Shape.getElement(sub).copy() for sub in subNames if sub.startswith("Edge")
            ]

        if edges:
            obj.Shape = Part.Wire(Part.__sortEdges__(edges))
        else:
            obj.Shape = Part.Shape()


class CommandPathShapeTC:
    def GetResources(self):
        return {
            "Pixmap": "CAM_ShapeTC",
            "MenuText": QT_TRANSLATE_NOOP("CAM_PathShapeTC", "Path From Shape TC"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "CAM_PathShapeTC",
                "Creates a path from the selected shapes with the tool controller",
            ),
        }

    def IsActive(self):
        isJob = False
        if FreeCAD.ActiveDocument is not None:
            for o in FreeCAD.ActiveDocument.Objects:
                if o.Name.startswith("Job"):
                    isJob = True
                    break
        if isJob:
            selection = FreeCADGui.Selection.getSelectionEx()
            if selection:
                base = selection[0].Object
                subBase = selection[0].SubElementNames
                if subBase and [edge for edge in subBase if "Edge" in edge]:
                    return True
                elif base.Shape.ShapeType in ["Wire", "Edge"]:
                    return True
        return False

    def Activated(self):
        print("Create PathShape object with Tool Controller")
        doc = FreeCAD.ActiveDocument
        selection = FreeCADGui.Selection.getSelectionEx()
        shapeObj = None
        if selection:
            base = selection[0].Object
            subBase = selection[0].SubElementNames
            if subBase:
                subEdges = [edge for edge in subBase if "Edge" in edge]
                shapeObj = doc.addObject("Part::FeaturePython", "PartShape")
                shapeObj.ViewObject.Proxy = 0
                shapeObj.Visibility = False
                shapeObj.Proxy = ObjectPartShape(shapeObj, [(base, subEdges)])
            elif base.Shape.ShapeType in ["Wire", "Edge"]:
                shapeObj = Draft.make_clone(base)

        pathObj = doc.addObject("Path::FeatureShape", "PathShape")
        pathObj.Sources = [shapeObj]

        # Overwrite getToolControllers() function with modified version
        PathUtils.getToolControllers = _getToolControllers

        PathUtils.addToJob(pathObj)
        _addBaseProperties(pathObj)
        _addToolController(pathObj)
        _setSafetyZ(pathObj)
        doc.recompute()


if FreeCAD.GuiUp:
    # Register the FreeCAD command
    FreeCADGui.addCommand("CAM_PathShapeTC", CommandPathShapeTC())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2013 WandererFan <wandererfan@gmail.com>                *
# *   Copyright (c) 2019 Zheng, Lei (realthunder)<realthunder.dev@gmail.com>*
# *   Copyright (c) 2020 Carlo Pavan <carlopav@gmail.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create PathArray objects.

The copies will be created along a path, like a polyline, B-spline,
or Bezier curve.
"""
## @package gui_patharray
# \ingroup draftguitools
# \brief Provides GUI tools to create PathArray objects.

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
import Draft_rc
import DraftVecUtils
import draftguitools.gui_base_original as gui_base_original

from draftutils.messages import _err
from draftutils.translate import translate

# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False


class PathArray(gui_base_original.Modifier):
    """Gui Command for the Path array tool.

    Parameters
    ----------
    use_link: bool, optional
        It defaults to `False`. If it is `True`, the created object
        will be a `Link array`.
    """

    def __init__(self, use_link=False):
        super(PathArray, self).__init__()
        self.use_link = use_link
        self.call = None

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_PathArray",
            "MenuText": QT_TRANSLATE_NOOP("Draft_PathArray", "Path Array"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_PathArray", "Creates copies of the selected object along a selected path"
            ),
        }

    def Activated(self, name="Path array"):
        """Execute when the command is called."""
        super(PathArray, self).Activated(name=name)
        self.name = name
        # This was deactivated because it doesn't work correctly;
        # the selection needs to be made on two objects, but currently
        # it only selects one.

        # if not Gui.Selection.getSelectionEx():
        #     if self.ui:
        #         self.ui.selectUi()
        #         _msg(translate("draft",
        #                        "Please select exactly two objects, "
        #                        "the base object and the path object, "
        #                        "before calling this command."))
        #         self.call = \
        #             self.view.addEventCallback("SoEvent",
        #                                        gui_tool_utils.selectObject)
        # else:
        #     self.proceed()
        self.proceed()

    def proceed(self):
        """Proceed with the command if one object was selected."""
        sel = Gui.Selection.getSelectionEx()
        if len(sel) != 2:
            _err(
                translate(
                    "draft",
                    "Select exactly 2 objects, the base object and the path object, before calling this command",
                )
            )
        else:
            base_object = sel[0].Object
            path_object = sel[1].Object

            count = 4
            extra = App.Vector(0, 0, 0)
            subelements = list(sel[1].SubElementNames)
            align = False
            align_mode = "Original"
            tan_vector = App.Vector(1, 0, 0)
            force_vertical = False
            vertical_vector = App.Vector(0, 0, 1)
            start_offset = 0.0
            end_offset = 0.0
            use_link = self.use_link

            _edge_list_str = list()
            _edge_list_str = ["'" + edge + "'" for edge in subelements]
            _sub_str = ", ".join(_edge_list_str)
            subelements_list_str = "[" + _sub_str + "]"

            vertical_vector_str = DraftVecUtils.toString(vertical_vector)

            Gui.addModule("Draft")
            _cmd = "Draft.make_path_array"
            _cmd += "("
            _cmd += "App.ActiveDocument." + base_object.Name + ", "
            _cmd += "App.ActiveDocument." + path_object.Name + ", "
            _cmd += "count=" + str(count) + ", "
            _cmd += "extra=" + DraftVecUtils.toString(extra) + ", "
            _cmd += "subelements=" + subelements_list_str + ", "
            _cmd += "align=" + str(align) + ", "
            _cmd += "align_mode=" + "'" + align_mode + "', "
            _cmd += "tan_vector=" + DraftVecUtils.toString(tan_vector) + ", "
            _cmd += "force_vertical=" + str(force_vertical) + ", "
            _cmd += "vertical_vector=" + vertical_vector_str + ", "
            _cmd += "start_offset=" + str(start_offset) + ", "
            _cmd += "end_offset=" + str(end_offset) + ", "
            _cmd += "use_link=" + str(use_link)
            _cmd += ")"

            _cmd_list = [
                "_obj_ = " + _cmd,
                "Draft.autogroup(_obj_)",
                "App.ActiveDocument.recompute()",
            ]
            self.commit(translate("draft", "Create Path Array"), _cmd_list)

        # Commit the transaction and execute the commands
        # through the parent class
        self.finish()


Gui.addCommand("Draft_PathArray", PathArray())


class PathLinkArray(PathArray):
    """Gui Command for the PathLinkArray tool based on the PathArray tool."""

    def __init__(self):
        super(PathLinkArray, self).__init__(use_link=True)

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_PathLinkArray",
            "MenuText": QT_TRANSLATE_NOOP("Draft_PathLinkArray", "Path Link Array"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_PathLinkArray",
                "Creates linked copies of the selected object along a selected path",
            ),
        }

    def Activated(self):
        """Execute when the command is called."""
        super(PathLinkArray, self).Activated(name="Path link array")


Gui.addCommand("Draft_PathLinkArray", PathLinkArray())

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# flake8: noqa E731
# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Widget for browsing Tool Library assets with filtering and sorting."""

import yaml
from typing import cast, List, Optional
from PySide import QtCore, QtGui
from PySide.QtGui import QMenu, QAction, QKeySequence
import FreeCAD
import Path
from ...assets import AssetManager, AssetUri
from ...toolbit import ToolBit
from ...toolbit.ui import ToolBitEditor
from ...toolbit.ui.util import natural_sort_key
from ...toolbit.ui.browser import ToolBitBrowserWidget, ToolBitUriRole
from ...toolbit.serializers import YamlToolBitSerializer
from ..models.library import Library


Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())
Path.Log.trackModule(Path.Log.thisModule())


class LibraryBrowserWidget(ToolBitBrowserWidget):
    """
    A widget to browse, filter, and select Tool Library assets from the
    AssetManager, with sorting and batch insertion, using a current library.
    """

    current_library_changed = QtCore.Signal()

    def __init__(
        self,
        asset_manager: AssetManager,
        store: str = "local",
        parent=None,
        compact=True,
    ):
        super().__init__(
            asset_manager=asset_manager,
            store=store,
            parent=parent,
            tool_no_factory=self.get_tool_no_from_current_library,
            compact=compact,
        )
        self.current_library: Optional[Library] = None
        self._selected_tool_type: Optional[str] = None
        self.layout().setContentsMargins(0, 0, 0, 0)

        # Add tool type filter combo box to the base widget
        self._tool_type_combo = QtGui.QComboBox()
        self._tool_type_combo.setSizePolicy(
            QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Preferred
        )
        self._top_layout.insertWidget(0, self._tool_type_combo, 1)
        self._tool_type_combo.currentTextChanged.connect(self._on_tool_type_combo_changed)

        self.restore_last_sort_order()
        self.load_last_library()

    def setDragEnabled(self, enabled: bool = True):
        """Enable or disable drag-and-drop support for the tool list."""
        self._tool_list_widget.setDragEnabled(enabled)

    def load_last_library(self):
        """Loads the last selected library from preferences."""
        library_uri = Path.Preferences.getLastToolLibrary()
        if library_uri:
            try:
                library = self._asset_manager.get(library_uri, store="local", depth=1)
                self.set_current_library(library)
            except Exception as e:
                Path.Log.warning(f"Failed to load last tool library: {e}")

    def restore_last_sort_order(self):
        """Sets the sort mode and updates the tool list."""
        last_sort_key = Path.Preferences.getLastToolLibrarySortKey()
        if last_sort_key:
            self.set_sort_order(last_sort_key)

    def set_sort_order(self, key: str):
        super().set_sort_order(key)
        Path.Preferences.setLastToolLibrarySortKey(self._sort_key)

    def _get_state(self):
        """Gets the current library URI, selected toolbit URI, and scroll
        position."""
        current_library_uri_str = (
            str(self.current_library.get_uri()) if self.current_library else None
        )

        selected_toolbit_uris = []
        selected_items = self._tool_list_widget.selectedItems()
        if selected_items:
            selected_toolbit_uris = [item.data(ToolBitUriRole) for item in selected_items]

        scroll_pos = self._tool_list_widget.verticalScrollBar().value()

        return {
            "library_uri": current_library_uri_str,
            "toolbit_uris": selected_toolbit_uris,
            "scroll_pos": scroll_pos,
        }

    def _set_state(self, selection_data):
        """Restores the library selection, toolbit selection, and scroll
        position."""
        library_uri_str = selection_data.get("library_uri")
        toolbit_uris = selection_data.get("toolbit_uris", [])
        scroll_pos = selection_data.get("scroll_pos", 0)

        # Restore library selection
        if library_uri_str:
            try:
                library_uri = AssetUri(library_uri_str)
                library = self._asset_manager.get(library_uri, store=self._store_name, depth=1)
                self.set_current_library(library)
            except FileNotFoundError:
                Path.Log.error(f"Library {library_uri_str} not found.")
                self.set_current_library(None)
        else:
            self.set_current_library(None)

        # Restore toolbit selection
        if toolbit_uris:
            for uri in toolbit_uris:
                for i in range(self._tool_list_widget.count()):
                    item = self._tool_list_widget.item(i)
                    if item.data(ToolBitUriRole) == uri:
                        item.setSelected(True)

        # Restore scroll position
        self._tool_list_widget.verticalScrollBar().setValue(scroll_pos)

    def refresh(self):
        """Refreshes the toolbits for the current library from disk."""
        Path.Log.debug("refresh(): Fetching and populating toolbits.")
        if self.current_library:
            library_uri = self.current_library.get_uri()
            try:
                self.current_library = cast(
                    Library, self._asset_manager.get(library_uri, store=self._store_name, depth=1)
                )
            except FileNotFoundError:
                Path.Log.error(f"Library {library_uri} not found.")
                self.current_library = None
        self._update_tool_list()

    def get_tool_no_from_current_library(self, toolbit):
        """
        Retrieves the tool number for a toolbit based on the current library.
        """
        if not self.current_library:
            return None
        tool_no = self.current_library.get_bit_no_from_bit(toolbit)
        return tool_no

    def set_current_library(self, library):
        """Sets the current library and updates the tool list."""
        self.current_library = library
        self._update_tool_list()
        self.current_library_changed.emit()

        # Save the selected library to preferences
        if library:
            Path.Preferences.setLastToolLibrary(str(library.get_uri()))

    def _get_available_tool_types(self):
        """Get all available tool types from the current assets."""
        tool_types = set()
        # Make sure we have assets to work with
        if not hasattr(self, "_all_assets") or not self._all_assets:
            return []

        for asset in self._all_assets:
            # Use get_shape_name() method to get the tool type
            if hasattr(asset, "get_shape_name"):
                tool_type = asset.get_shape_name()
                if tool_type:
                    tool_types.add(tool_type)

        return sorted(tool_types)

    def _get_filtered_assets(self):
        """Get assets filtered by tool type if a specific type is selected."""
        if self._tool_type_combo.currentIndex() == 0:  # "All Toolbit Types"
            return self._all_assets

        filtered_assets = []
        for asset in self._all_assets:
            if hasattr(asset, "get_shape_name"):
                tool_type = asset.get_shape_name()
                if tool_type == self._selected_tool_type:
                    filtered_assets.append(asset)
        return filtered_assets

    def _update_tool_list(self):
        """Updates the tool list based on the current library."""
        if self.current_library:
            self._all_assets = [t for t in self.current_library]
        else:
            # Fetch all toolbits
            all_toolbits = self._asset_manager.fetch(asset_type="toolbit", depth=0)
            self._all_assets = cast(List[ToolBit], all_toolbits)
        self._sort_assets()
        self._tool_list_widget.clear_list()
        # Update tool type combo after assets are loaded
        if hasattr(self, "_tool_type_combo"):
            self._update_tool_type_combo()
        self._update_list()

    def _update_list(self):
        """Updates the list widget with filtered assets."""
        self._tool_list_widget.clear_list()
        filtered_assets = self._get_filtered_assets()

        # Apply search filter if there is one
        search_term = self._search_edit.text().lower()
        if search_term:
            search_filtered = []
            for asset in filtered_assets:
                if search_term in asset.label.lower():
                    search_filtered.append(asset)
                    continue
                # Also search in tool type
                if hasattr(asset, "get_shape_name"):
                    tool_type = asset.get_shape_name()
                    if tool_type and search_term in tool_type.lower():
                        search_filtered.append(asset)
            filtered_assets = search_filtered

        for asset in filtered_assets:
            self._tool_list_widget.add_toolbit(asset)

    def _add_shortcuts(self):
        """Adds keyboard shortcuts for common actions."""
        Path.Log.debug("LibraryBrowserWidget._add_shortcuts: Called.")
        super()._add_shortcuts()

        cut_action = QAction(self)
        cut_action.setShortcuts(QKeySequence.Cut)
        cut_action.triggered.connect(self._on_cut_requested)
        self.addAction(cut_action)

        duplicate_action = QAction(self)
        duplicate_action.setShortcut(QKeySequence("Ctrl+D"))
        duplicate_action.triggered.connect(self._on_duplicate_requested)
        self.addAction(duplicate_action)

        remove_action = QAction(self)
        remove_action.setShortcut(QKeySequence.Delete)
        remove_action.triggered.connect(self._on_remove_from_library_requested)
        self.addAction(remove_action)

        paste_action = QAction(self)
        paste_action.setShortcuts(QKeySequence.Paste)
        paste_action.triggered.connect(self._on_paste_requested)
        self.addAction(paste_action)

    def _show_context_menu(self, position):
        """Shows the context menu at the given position."""
        context_menu = QMenu(self)

        selected_items = self._tool_list_widget.selectedItems()
        has_selection = bool(selected_items)
        has_library = self.current_library is not None

        # Add actions in the desired order
        edit_action = context_menu.addAction("Edit", self._on_edit_requested)
        edit_action.setEnabled(has_selection)

        context_menu.addSeparator()

        action = context_menu.addAction("Copy", self._on_copy_requested)
        action.setShortcut(QtGui.QKeySequence("Ctrl+C"))

        action = context_menu.addAction("Cut", self._on_cut_requested)
        action.setShortcut(QtGui.QKeySequence("Ctrl+X"))

        action = context_menu.addAction("Paste", self._on_paste_requested)
        action.setShortcut(QtGui.QKeySequence("Ctrl+V"))

        # Paste is enabled if there is data in the clipboard
        clipboard = QtGui.QApplication.clipboard()
        mime_type = "application/x-freecad-toolbit-list-yaml"
        action.setEnabled(clipboard.mimeData().hasFormat(mime_type))

        action = context_menu.addAction("Duplicate", self._on_duplicate_requested)
        action.setShortcut(QtGui.QKeySequence("Ctrl+D"))

        context_menu.addSeparator()

        # Only show "Remove from Library" when viewing a specific library
        if has_library:
            action = context_menu.addAction(
                "Remove from Library", self._on_remove_from_library_requested
            )
            action.setShortcut(QtGui.QKeySequence.Delete)

        # Only show "Delete from disk" when viewing 'all tools' (no library selected)
        if not has_library:
            action = context_menu.addAction("Delete from disk", self._on_delete_requested)
            action.setShortcut(QtGui.QKeySequence("Shift+Delete"))

        # Execute the menu
        context_menu.exec_(self._tool_list_widget.mapToGlobal(position))

    def get_current_library(self) -> Library | None:
        """Helper to get the current library."""
        return self.current_library

    def _on_edit_requested(self):
        """Opens the ToolBitEditor for the selected toolbit."""
        toolbit = self._get_first_selected_bit()
        if not toolbit:
            return

        # Open the editor for the selected toolbit
        tool_no = self.get_tool_no_from_current_library(toolbit)
        editor = ToolBitEditor(toolbit, tool_no, parent=self)
        result = editor.show()
        if result != QtGui.QDialog.Accepted:
            return

        # If the editor was closed with "OK", save the changes
        self._asset_manager.add(toolbit)
        Path.Log.info(f"Toolbit {toolbit.get_id()} saved.")

        # Also save the library because the tool number may have changed.
        if self.current_library and tool_no != editor.tool_no:
            self.current_library.assign_new_bit_no(toolbit, editor.tool_no)
            self._asset_manager.add(self.current_library)

        state = self._get_state()
        self.refresh()
        self._update_list()
        self._set_state(state)

    def _on_cut_requested(self):
        """Handles cut request by copying and marking for removal from library."""
        uris = self.get_selected_bit_uris()
        library = self.get_current_library()
        if not library or not uris:
            return

        # Copy to clipboard (handled by base class _to_clipboard)
        extra_data = {"source_library_uri": str(library.get_uri())}
        self._to_clipboard(uris, mode="cut", extra_data=extra_data)

    def _on_duplicate_requested(self):
        """Handles duplicate request by duplicating and adding to library."""
        Path.Log.debug("LibraryBrowserWidget._on_duplicate_requested: Called.\n")
        uris = self.get_selected_bit_uris()
        library = self.get_current_library()
        if not library or not uris:
            Path.Log.debug(
                "LibraryBrowserWidget._on_duplicate_requested: No library or URIs selected. Returning."
            )
            return

        new_uris = set()
        for uri_string in uris:
            toolbit = cast(ToolBit, self._asset_manager.get(AssetUri(uri_string), depth=0))
            if not toolbit:
                Path.Log.warning(f"Toolbit {uri_string} not found.\n")
                continue

            # Change the ID of the toolbit and save it to disk
            toolbit.set_id()  # Generate a new ID
            toolbit.label = toolbit.label + " (copy)"
            added_uri = self._asset_manager.add(toolbit)
            if added_uri:
                new_uris.add(str(toolbit.get_uri()))

            # Add the bit to the current library
            library.add_bit(toolbit)

        self._asset_manager.add(library)  # Save the modified library
        self.refresh()

        self.select_by_uri(list(new_uris))

    def _on_paste_requested(self):
        """Handles paste request by adding toolbits to the current library."""
        current_library = self.get_current_library()
        if not current_library:
            return

        clipboard = QtGui.QApplication.clipboard()
        mime_type = "application/x-freecad-toolbit-list-yaml"
        mime_data = clipboard.mimeData()

        if not mime_data.hasFormat(mime_type):
            return

        try:
            clipboard_content_yaml = mime_data.data(mime_type).data().decode("utf-8")
            clipboard_data_dict = yaml.safe_load(clipboard_content_yaml)

            if (
                not isinstance(clipboard_data_dict, dict)
                or "toolbits" not in clipboard_data_dict
                or not isinstance(clipboard_data_dict["toolbits"], list)
            ):
                return

            serialized_toolbits_data = clipboard_data_dict["toolbits"]
            mode = clipboard_data_dict.get("operation", "copy")
            source_library_uri_str = clipboard_data_dict.get("source_library_uri")

            if mode == "copy":
                self._on_copy_paste(current_library, serialized_toolbits_data)
            elif mode == "cut" and source_library_uri_str:
                self._on_cut_paste(
                    current_library, serialized_toolbits_data, source_library_uri_str
                )

        except Exception as e:
            Path.Log.warning(f"An unexpected error occurred during paste: {e}")

    def _on_copy_paste(self, current_library: Library, serialized_toolbits_data: list):
        """Handles pasting toolbits that were copied."""
        new_uris = set()
        for toolbit_yaml_str in serialized_toolbits_data:
            if not isinstance(toolbit_yaml_str, str) or not toolbit_yaml_str.strip():
                continue

            toolbit_data_bytes = toolbit_yaml_str.encode("utf-8")
            toolbit = YamlToolBitSerializer.deserialize(toolbit_data_bytes, dependencies=None)

            # Get the original toolbit ID from the deserialized data
            original_id = toolbit.id
            Path.Log.info(f"COPY PASTE: Attempting to paste toolbit with original_id={original_id}")

            # Check if toolbit already exists in asset manager
            toolbit_uri = toolbit.get_uri()
            existing_toolbit = None
            try:
                existing_toolbit = self._asset_manager.get(
                    toolbit_uri, store=["local", "builtin"], depth=0
                )
                Path.Log.info(f"COPY PASTE: Found existing toolbit {original_id}, using reference")
            except FileNotFoundError:
                # Toolbit doesn't exist, save it as new
                Path.Log.info(f"COPY PASTE: Toolbit {original_id} not found, creating new one")
                self._asset_manager.add(toolbit)
                existing_toolbit = toolbit

            # Add the existing or new toolbit to the current library
            added_toolbit = current_library.add_bit(existing_toolbit)
            if added_toolbit:
                new_uris.add(str(existing_toolbit.get_uri()))

        if new_uris:
            self._asset_manager.add(current_library)  # Save the modified library
            self.refresh()
            self.select_by_uri(list(new_uris))

    def _on_cut_paste(
        self,
        current_library: Library,
        serialized_toolbits_data: list,
        source_library_uri_str: str,
    ):
        """Handles pasting toolbits that were cut."""
        source_library_uri = AssetUri(source_library_uri_str)
        if source_library_uri == current_library.get_uri():
            # Cut from the same library, do nothing
            return

        try:
            source_library = cast(
                Library,
                self._asset_manager.get(source_library_uri, store=self._store_name, depth=1),
            )
        except FileNotFoundError:
            Path.Log.warning(f"Source library {source_library_uri_str} not found.\n")
            return

        new_uris = set()
        for toolbit_yaml_str in serialized_toolbits_data:
            if not isinstance(toolbit_yaml_str, str) or not toolbit_yaml_str.strip():
                continue

            toolbit_data_bytes = toolbit_yaml_str.encode("utf-8")
            toolbit = YamlToolBitSerializer.deserialize(toolbit_data_bytes, dependencies=None)

            # Get the original toolbit ID and find the existing toolbit
            original_id = toolbit.id
            Path.Log.info(f"CUT PASTE: Moving toolbit with original_id={original_id}")

            toolbit_uri = toolbit.get_uri()
            try:
                existing_toolbit = self._asset_manager.get(
                    toolbit_uri, store=["local", "builtin"], depth=0
                )
                Path.Log.info(f"CUT PASTE: Found existing toolbit {original_id}, using reference")

                # Remove from source library, add to target library
                source_library.remove_bit(existing_toolbit)
                added_toolbit = current_library.add_bit(existing_toolbit)
                if added_toolbit:
                    new_uris.add(str(existing_toolbit.get_uri()))
            except FileNotFoundError:
                Path.Log.warning(f"CUT PASTE: Toolbit {original_id} not found in asset manager")

        if new_uris:
            # Save the modified libraries
            self._asset_manager.add(current_library)
            self._asset_manager.add(source_library)
            self.refresh()
            self.select_by_uri(list(new_uris))

    def _on_remove_from_library_requested(self):
        """Handles request to remove selected toolbits from the current library."""
        Path.Log.debug("_on_remove_from_library_requested: Called.")
        uris = self.get_selected_bit_uris()
        library = self.get_current_library()
        if not library or not uris:
            return

        # Ask for confirmation
        reply = QtGui.QMessageBox.question(
            self,
            FreeCAD.Qt.translate("CAM", "Confirm Removal"),
            FreeCAD.Qt.translate(
                "CAM", "Are you sure you want to remove the selected toolbit(s) from the library?"
            ),
            QtGui.QMessageBox.Yes | QtGui.QMessageBox.No,
            QtGui.QMessageBox.No,
        )

        if reply == QtGui.QMessageBox.Yes:
            self._remove_toolbits_from_library(library, uris)

    def _remove_toolbits_from_library(self, library: Library, uris: List[str]):
        """Removes toolbits with the given URIs from the specified library."""
        removed_count = 0
        for uri_string in uris:
            try:
                # Remove the toolbit from the library
                library.remove_bit_by_uri(uri_string)
                removed_count += 1
            except Exception as e:
                Path.Log.error(f"Failed to remove toolbit {uri_string} from library: {e}\n")

        if removed_count > 0:
            self._asset_manager.add(library)
            self.refresh()

    def _update_tool_type_combo(self):
        """Update the tool type combo box with available types."""
        current_selection = self._tool_type_combo.currentText()
        self._tool_type_combo.blockSignals(True)
        try:
            self._tool_type_combo.clear()
            self._tool_type_combo.addItem(FreeCAD.Qt.translate("CAM", "All Toolbit Types"))

            for tool_type in self._get_available_tool_types():
                self._tool_type_combo.addItem(tool_type)

            # Restore selection if it still exists
            index = self._tool_type_combo.findText(current_selection)
            if index >= 0:
                self._tool_type_combo.setCurrentIndex(index)
            else:
                self._tool_type_combo.setCurrentIndex(0)
        finally:
            self._tool_type_combo.blockSignals(False)

    def _on_tool_type_combo_changed(self, tool_type):
        """Handle tool type filter selection change."""
        self._selected_tool_type = tool_type
        self._update_list()


class LibraryBrowserWithCombo(LibraryBrowserWidget):
    """
    A widget extending LibraryBrowserWidget with a combo box for library selection.
    """

    def __init__(
        self,
        asset_manager: AssetManager,
        store: str = "local",
        parent=None,
        compact=True,
    ):
        super().__init__(
            asset_manager=asset_manager,
            store=store,
            parent=parent,
            compact=compact,
        )

        # Move search box into dedicated row to make space for the
        # library selection combo box
        layout = self.layout()
        self._top_layout.removeWidget(self._search_edit)
        layout.insertWidget(1, self._search_edit, 20)

        # Library selection combo box
        self._library_combo = QtGui.QComboBox()
        self._library_combo.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Preferred)
        self._top_layout.insertWidget(0, self._library_combo, 1)
        self._library_combo.currentIndexChanged.connect(self._on_library_combo_changed)

        self._top_layout.removeWidget(self._tool_type_combo)
        self._top_layout.insertWidget(1, self._tool_type_combo, 1)

        self.current_library_changed.connect(self._on_current_library_changed)

        self._in_refresh = False
        self.refresh()

    def _on_library_combo_changed(self, index):
        """Handles library selection change from the combo box."""
        if self._in_refresh:
            return

        selected_library = cast(Library, self._library_combo.itemData(index))
        if not selected_library:
            return

        # Have to refetch the non-shallow library.
        uri = selected_library.get_uri()
        library = self._asset_manager.get(uri, store=self._store_name, depth=1)
        self.set_current_library(library)

    def _on_current_library_changed(self):
        """Updates the combo box when the current library changes externally."""
        if self.current_library:
            for i in range(self._library_combo.count()):
                lib = self._library_combo.itemData(i)
                if lib.get_uri() == self.current_library.get_uri():
                    self._library_combo.setCurrentIndex(i)
                    return
            Path.Log.warning(
                f"Current library {self.current_library.get_uri()} not found in combo box."
            )

    def refresh(self):
        """Reads available libraries and refreshes the combo box and toolbits."""
        Path.Log.debug("refresh(): Fetching and populating libraries and toolbits.")
        libraries = self._asset_manager.fetch("toolbitlibrary", store=self._store_name, depth=0)
        self._in_refresh = True
        try:
            self._library_combo.clear()
            for library in sorted(libraries, key=lambda x: natural_sort_key(x.label)):
                self._library_combo.addItem(library.label, userData=library)
        finally:
            self._in_refresh = False

        super().refresh()

        if not libraries:
            return
        if not self.current_library:
            first_library = self._library_combo.itemData(0)
            if first_library:
                uri = first_library.get_uri()
                library = self._asset_manager.get(uri, store=self._store_name, depth=1)
                self.set_current_library(library)
            self._library_combo.setCurrentIndex(0)
            return

        for i in range(self._library_combo.count()):
            lib = self._library_combo.itemData(i)
            if lib.get_uri() == self.current_library.get_uri():
                self._library_combo.setCurrentIndex(i)
                break
        else:
            self._library_combo.setCurrentIndex(0)
